#
# Unit tests for the circuits package
#


from fixture_circuits import circuit_constant2acl_out

def test_constant_circuit():
    carrier = circuit_constant2acl_out().generate()
    print(carrier)
    assert "/0" in carrier
    assert "constant" in carrier["/0"]["/U"]
    assert carrier["/0"]["/U"]["constant"] == True
    



#!/usr/bin/env python3

"""
An Synchronous Hybrid Controller Python Client for REDAC/LUCIDAC.

This is a minimal python client with focus on interactive python prompt usage.

This client implementation does *not* feature strong typing, dataclasses,
asynchronous functions. Instead, it implements a blocking API and tries to
mimic the way how the *Model-1 Hybrid Controller* interface worked
(in particular in the `PyAnalog <https://github.com/anabrid/pyanalog/>`_
research code).

This is a single file implementation focussing on portability and minimal
dependencies. If you have ``pyserial`` installed, it will be used to allow
serial connections, otherwise it runs with python "batteries included" to
connect to TCP/IP targets.

In particular, there are no dependencies on any other lucipy module, except
:mod:`~lucipy.detect`. In particular, this code is completely independent from
the :mod:`~lucipy.circuits` module which can serve to simplify to set up the
data structure required by :meth:`LUCIDAC.set_circuit`.

The main class to use from this module is :class:`LUCIDAC`.
"""

# all this is only python standard library  :)
import logging, time, socket, select, json, types, typing, \
    itertools, os, functools, collections, uuid, time
log = logging.getLogger('synchc')
logging.basicConfig(level=logging.INFO)

from .detect import detect, Endpoint

__all__ = """
    LUCIDAC Run LUCIGroup
    RemoteError LocalError
""".split()

nonempty = lambda lst: [x for x in lst if x]

try:
    import serial
except ModuleNotFoundError:
    serial = None

class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    #__getattr__ = dict.get
    def __getattr__(*args):
        val = dict.__getitem__(*args)
        return dotdict(val) if type(val) is dict else val
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

def has_data(fh):
    "Peeks a file handle (checks for data without reading/consuming)"
    rlist, wlist, xlist = select.select([fh], [],[], 0)
    return len(rlist) != 0


class RemoteError(Exception):
    """
    A RemotError represents a returned error message by the endpoint.
    Typically this means the LUCIDAC HybridController could not decode a message,
    there is some typing or logic problem.
    
    The error message from the server always contains an integer error code and
    a human readable string error message.
    
    In order to trace the error at the firmware side, both the code and string
    can be helpful. You can grep the firwmare code with the string. For the code,
    you have to understand how they are computed server side or consult the
    firmware documentation.
    """
    def __init__(self, recv_envelope):
        self.code = recv_envelope.code
        self.type = recv_envelope.type
        self.raw = recv_envelope
        super().__init__(f"Remote error {recv_envelope.code} at query {recv_envelope.type}: '{recv_envelope.error}'")


class LocalError(ValueError):
    """
    A LocalError represents a logic flaw at the client side, i.e. within the lucipy
    code. It did not expect some message from the server or could not deserialize
    messages received.
    
    Note that next to LocalErrors you can always receive things such as SocketError,
    OSError or similiar low level connection problems (for instance: OSError "no route
    to host")
    """
    pass

class tcpsocket:
    "A socket with readline support"
    def __init__(self, host, port, auto_reconnect=True):
        self.host, self.port, self.auto_reconnect = host, port, auto_reconnect
        self.debug_print = False
        self.connect()
    def connect(self):
        if(hasattr(self, 's')):
            log.warning(f"Trying to reconnect to {self.host}:{self.port}...")
        else:
            log.info(f"Connecting to TCP {self.host}:{self.port}...")
        self.s = socket.socket()
        self.s.connect((self.host,self.port))
        self.fh = self.s.makefile(mode="rw", encoding="utf-8")
    def close(self):
        return self.s.close()
        #del self.s
    def send(self, sth):
        "Expects sth to be a string"
        try:
            if self.debug_print:
                print(f"tcpsocket.send({sth=})")
            #self.s.sendall(sth.encode("ascii"))
            self.fh.write(sth + "\n")
            self.fh.flush()
            #print("tcpsocket.send() completed")
        except (BrokenPipeError, ConnectionResetError) as e:
            if self.debug_print:
                print(f"tcpsocket.send: {e}")
            if self.auto_reconnect:
                self.connect()
                return self.send(sth)
            else:
                raise e
    def read(self, *args, **kwargs):
        "Returns a complete line as string. See instead also: self.s.recv(123)"
        try:
            #print("tcpsocket.readline()")
            #import ipdb; ipdb.set_trace()
            line = self.fh.readline()
            if self.debug_print:
                print(f"tcpsocket.read() = {line}")
            return line
        except UnicodeDecodeError as e:
            print(e)
            return ""
        except ConnectionResetError as e:
            if self.debug_print:
                print(f"tcpsocket.read: {e}")
            if self.auto_reconnect:
                self.connect()
                return "" # empty line, since query protcol should not readline empty socket
            else:
                raise e
    def has_data(self):
        return has_data(self.s)
    def __repr__(self):
        return f"tcp://{self.host}:{self.port}"
    
# TODO: Probably refactor code, the classes tcpsocket and serialsocket share most of their logic

class serialsocket:
    "Uses pyserial to connect to directly attached device"
    def __init__(self, device):
        if not serial:
            raise ImportError("PySerial not available, please install with 'pip install pyserial'")
        self.device = device
        self.fh = serial.Serial(self.device)
        # sometimes there is stuff stuck in the serial port. Read all of it and wipe it.
        while self.has_data():
            self.fh.readline()
    def close(self):
        return self.fh.close()
    def send(self, sth):
        #print(f"serialsocket.send({sth})")
        self.fh.write(sth.encode("ascii") + b"\n")
        self.fh.flush()
        #print(f"serialsocket.send completed")
    def read(self):
        # block until have read exactly one line
        while self.has_data():
            ret = self.fh.readline()
            #print(f"serialsocket.read(): {ret}")
            return ret
    def has_data(self):
        return has_data(self.fh)
    def __repr__(self):
        return f"socket:/{self.device}"

class emusocket:
    "Emulates a socket with a callback function"
    def __init__(self, callback=None, debug=False):
        if not callback:
            from .simulator import Emulation
            emu = Emulation(debug=debug)
            callback = lambda line: emu.handle_request(line)
        self.callback = callback
        self.return_buffer = []
    def send(self, sth):
        ret = self.callback(sth)
        if isinstance(ret, list):
            self.return_buffer += ret
        else:
            self.return_buffer.append(ret)
    def read(self):
        if self.has_data():
            return self.return_buffer.pop(0)
    def close(self):
        pass
    def has_data(self):
        return len(self.return_buffer)
    def __repr__(self):
        return f"emu:/?callback={self.callback}"

class jsonlines:
    "Middleware that speaks dictionaries at front and JSON at back"
    def __init__(self, actual_socket):
        self.sock = actual_socket
    @staticmethod
    def makeSocket(cls, actual_socket_type, *args, **kwargs):
        return cls(actual_socket_type(*args, **kwargs))
    def send(self, sth):
        #print(f"jsonlines.send({json.dumps(sth)}")
        self.sock.send(json.dumps(sth))
        #print(f"jsonlines.send completed")
    def read(self, *args, **kwargs):
        #print("jsonlines.read()")
        read = self.sock.read(*args, **kwargs)
        while not read:
            print("haven't read anything, trying again")
            time.sleep(0.2)
            read = self.sock.read(*args, **kwargs)
        #print(f"jsonlines.read() got {read}")
        try:
            return json.loads(read)
        except json.JSONDecodeError as s:
            raise LocalError(f"Could not decode as JSON: {s}. Received message ({len(read)} characters) is '{read}'")
    def close(self):
        return self.sock.close()
    def read_all(self):
        while self.sock.has_data():
            yield self.read()

def endpoint2socket(endpoint_url: typing.Union[Endpoint,str]) -> typing.Union[tcpsocket,serialsocket]:
    "Provides the appropriate *synchronous* socket for a given endpoint"
    endpoint = Endpoint(endpoint_url)
    if endpoint.scheme == "serial": # serial:/dev/ttyFooBar
        return serialsocket(endpoint.host)
    elif endpoint.scheme == "tcp": # tcp://192.168.1.2:5732
        return tcpsocket(endpoint.host, endpoint.port, auto_reconnect="auto_reconnect" in endpoint.args)
    elif endpoint.scheme in ["emu","sim"]: # emu:/ or emu:/?debug
        return emusocket(debug="debug" in endpoint.args)
    elif endpoint.scheme == "zeroconf":
        endpoint_url = detect(single=True)
        if not endpoint_url:
            raise ValueError("zeroconf:/ explicitely asked for, but no endpoint provided as argument or in ENV variable and could also not discover something on USB or in Network.")
        return endpoint2socket(endpoint_url)
    else:
        raise ValueError(f"Illegal {endpoint_url=}. Expecting something like tcp://192.168.1.2:5732 or serial:/dev/foo")


class Run:
    """
    A run is a IC-OP-HALT sequence on a LUCIDAC with measurement (data aquisition with the analog-digital-converters)
    ongoing during the OP phase. Therefore, a run generates data. This results in the remote site to send data "out of bounds",
    i.e. without that we have requested these data. The job of this class is to proper model how to receive these data.
    An instance of this class is returned by :meth:`LUCIDAC.start_run`. This instance is properly "handled off" by
    calling :meth:`data` and then wiping it.
    
    """
    run_states = "DONE ERROR IC NEW OP OP_END QUEUED TAKE_OFF TMP_HALT".split()
    
    def __init__(self, hc):
        self.hc = hc
    
    def next_data(self) -> typing.Iterator[typing.List[float]]:
        """
        Reads next dataset from DAQ (data aquisiton) which is streamed during the run.
        A call to this function yields a single dataset once arrived. It returns
        when the run is finished and raises a LocalError if it doesn't properly
        stop.
    
        The shape of data returned by this call is determined by the requested number
        of DAQ channels, which is between 0 (= no data will be returned at all, not
        even an empty directory per dataset) and 8 (= all channels). That is, the
        following invariant holds:
    
        >>> lines = run.next_data()                                                    # doctest: +SKIP
        >>> assert all(hc.daq_config["num_channels"] == len(line) for line in lines)   # doctest: +SKIP
    
        This invariant is also asserted within the method.
        """
        while True:
            envelope = self.hc.sock.read()
            if envelope["type"] == "run_data":
                # TODO check for proper run id and entity.
                msg_data = envelope["msg"]["data"]
                assert all(self.hc.daq_config["num_channels"] == len(line) for line in msg_data)
                yield msg_data
            elif envelope["type"] == "run_state_change":
                msg_old = envelope["msg"]["new"]
                msg_new = envelope["msg"]["new"]
                if msg_new == "DONE":
                    break # successfully transfered all data
                if msg_new == "ERROR":
                    # TODO: This is actually a RemoteError,
                    #  but since we don't get a proper envelope but some weird extra message
                    #  which does not even hold a suitable error string, we have to come up
                    #  with something ourselves...
                    raise LocalError(f"Could not properly start the run. Most likely the DAQ ({self.hc.daq_config}) or RUN ({self.hc.run_config}) configuration not accepted by the server side.")
                # Attention: After runstate stop there still can come a last data
                #            package.
            else:
                raise LocalError(f"Run::slurp(): Unexpected message {envelope}")
                #return # stop slurping

    def data(self, empty_is_fine=False) -> typing.List[float]:
        """
        Returns all measurement data (evolution data on the ADCs) during a run.
        
        This is basically a synchronous wait until the run finishes.
        
        The shape of data returned by this call is basically
        ``NUM_SAMPLING_POINTS x NUM_CHANNELS``. Since this is a uniform array, users
        can easily use numpy to postprocess and extract what they are interested in.
        Typically, you want to trace the evolution of particular channels, for instance
        
        >>> data = np.array(run.data())                      # doctest: +SKIP
        >>> x, y, z = data[:,0], data[:,1], data[:,2]        # doctest: +SKIP
        >>> plt.plot(x)                                      # doctest: +SKIP
    
        See also :meth:`next_data` and example application codes.
        
        :arg empty_is_fine: Whether to raise when no data have been aquired or
           happily return an empty array. Raising a ``LocalError`` (i.e. the default
           option) is most likely more what you want because you then don't have to
           write error handling code for an empty array. Otherwise a later access on
           something on  ``np.array(run.data())`` will most likely result in an
           ``IndexError: index 0 is out of bounds for axis 0 with size 0`` or similar.
        """
        res = sum(self.next_data(), []) # joins lists at outer level
        if len(res) == 0 and not empty_is_fine:
            raise LocalError("Expected data stream but got not a single data point")
        return res

class LUCIDAC:
    """
    This kind of class is known as *HybridController* in other codes. It serves as
    primary entry point for the lucipy code.
        
    The constructor has a variety of variants how to be called:
        
    :param endpoint_url: If no endpoint (either as string or Endpoint class instance)
        is provided, will lookup the environment variable ``LUCIDAC_ENDPOINT``. 
        
        If neither an endpoint nor the environment variable is set, autodetection
        is applied and the first connection is chosen. Note that if no LUCIDAC
        is attached via USB serial, the zeroconf detection will require a few
        hundred milliseconds, depending on your network.
        
        For details, see :ref:`lucipy-detection`.
    :param auto_reconnect: Whether reconnect in case of loss connection (TODO: Move
        to parameter ``?reconnect`` in the endpoint URL syntax)
    """
   
    ENDPOINT_ENV_NAME = "LUCIDAC_ENDPOINT"
    
    # a list of commands which will be exposed as methods, for shorthands.

    # TODO: Maybe write more verbose as in pyanalog in order to have a more usable documentation!
    commands = """
        ping help
        reset_circuit set_circuit get_circuit
        get_entities
        stop_run
        one_shot_daq
        manual_mode
        net_get net_set net_reset net_status
        login
        lock_acquire lock_release
        sys_ident sys_reboot
    """.split()
    
    # Commands which can be memoized for a given endpoint/instance, makes
    # it cheaper to call them repeatedly
    memoizable = "sys_ident".split()
    ## TODO: Get rid of this memoization because with arguments it can lead to weird errors, for instance
    #    hc.sys_ident(dict(blink=True))  -> TypeError: unhashable type: 'dict'
    
    def __init__(self, endpoint_url=None, auto_reconnect=True):
        if not endpoint_url:
            if self.ENDPOINT_ENV_NAME in os.environ:
                endpoint_url = os.environ[self.ENDPOINT_ENV_NAME]
            else:
                endpoint_url = detect(single=True) 
                # should raise an ModuleNotFoundError if and only if a library is missing which could have found something
                if not endpoint_url:
                    raise ValueError("No endpoint provided as argument or in ENV variable "
                                     + self.ENDPOINT_ENV_NAME + 
                                     " and did not discover an USB or network endpoint. No missing external libraries encountered.")
                    # self.ENDPOINT_ENV_NAME is used instead of the hardcoded string in case
                    # one wants to modify the env variable name.

        endpoint = Endpoint(endpoint_url)
        socket = endpoint2socket(endpoint_url)
        self.sock = jsonlines(socket)
        self.req_id = 50
        
        #: Ethernet Mac address of Microcontroller, required for the circuit entity hierarchy
        self.hc_mac = None
        # TODO: Maybe change firmware once in a way that it doesn't really need the client to know it
        
        #: Storage for stateful preparation of runs.
        self.run_config = dotdict(
            halt_on_external_trigger = False,
            halt_on_overload  = True,
            ic_time = 123456, # ns 
            op_time = 234567, # ns
        )

        #: Storage for stateful preparation of runs.
        self.daq_config = dotdict(
            num_channels = 0,
            sample_op = True,
            sample_op_end = True,
            sample_rate = 500_000,
        )
        
        #: Storage for additional configuration options which go next to each set_circuit call
        self.circuit_options = dotdict(
            # -> should all be moved into the firmware!
            #reset_before=True,
            #sh_kludge=True,
            #mul_calib_kludge=True,
            #calibrate_mblock=False,
            #calibrate_offset=False,
            #calibrate_routes=False,
        )
        
        self.user = endpoint.user
        self.password = endpoint.password
        
        if self.user:
            self.login()
    
    def __repr__(self):
        return f"LUCIDAC(\"{self.sock.sock}\")"
    
    def close(self):
        """
        Closes connection to LUCIDAC. This will close the socket and should be used
        when destructing the object. In most scripts it should be not neccessary to call
        ``close`` explicitely becaue the connection is considered ephermal for the script
        runtime. However, in unit testing conditions it might be interesting to close a
        connection in order to be able to reuse the socket ports later.
        """
        return self.sock.close()
    
    def send(self, msg_type, msg={}):
        "Sets up an envelope and sends it, but does not wait for reply"
        envelope = dict(id=self.req_id, type=msg_type, msg=msg)
        self.req_id += 1
        self.sock.send(envelope)
        return envelope
    
    def _recv(self, sent_envelope, ignore_run_state_change=True):
        resp = dotdict(self.sock.read())
        if ignore_run_state_change and "type" in resp and resp.type == "run_state_change":
            log.info(f"run_state_change: {resp.msg}")
            return self._recv(sent_envelope, ignore_run_state_change=True)
        if "type" in resp and resp.type == "log":
            log.info(f"Device Log {resp}")#[{resp.time}] {resp.msg}")
            return self._recv(sent_envelope, ignore_run_state_change=True)
        if sent_envelope == resp:
            # This is a serial socket replying first what was typed. Read another time.
            return self._recv(sent_envelope, ignore_run_state_change=ignore_run_state_change)
        if "error" in resp:
            raise RemoteError(resp)
        if resp.type == sent_envelope.type:
            # Do not show the empty message, in case of success.
            return resp.msg if resp.msg != {} else None
        else:
            log.error(f"req(type={sent_envelope.type}) received unexpected: {resp=}")
            return resp

    def query(self, msg_type , msg={}, ignore_run_state_change=True):
        "Sends a query and waits for the answer, returns that answer"
        envelope = dotdict(self.send(msg_type, msg))
        return self._recv(envelope)
    
    def slurp(self):
        """
        Flushes the input buffer in the socket.
        
        Returns all messages read as a list. This is useful when we lost the remote
        state (wrongly tracked) and have to clean up.
    
        .. note::
        
           Calling this method should not be possible in regular usage. In particular, TCP
           connections are normally "clean". In contrast, serial terminals are ephermal
           from the Microcontrollers point of view, so it might be neccessary to flush
           the connection in particular at beginning of a connection.
        """
        return list(self.sock.read_all())
    
    def login(self, user=None, password=None):
        """
        Login to the system. If no credentials are given, falls back to the one from the
        endpoint URL.
        """
        if user == None:
            user = self.user
        if password == None:
            password = self.password
        if not user:
            raise ClientError(f"Cannot login because no user name was given. Endpoint is {this.endpoint}")
        # won't check on empty password despite it also is basically an error
        return self.query("login", dict(user=user, password=password))
        
    
    def get_mac(self):
        "Get system ethernet mac address. Is cached."
        if not self.hc_mac:
            self.get_entities()
        return self.hc_mac
    
    def get_entities(self):
        "Gets entities and determines system Mac address"
        entities = self.query("get_entities")["entities"]
        mac = list(entities.keys())[0]
        # todo, assert that all this went fine, i.e. no KeyErrors or similar
        if mac[0] == "/": # if entity begins like "/04-E9-E5", wipe "/"
            mac = mac[1:]        
        assert self.hc_mac == None or mac == self.hc_mac, "Inconsistent device mac"
        self.hc_mac = mac
        return entities

    @staticmethod
    def determine_idal_ic_time_from_k0s(mIntConfig):
        """
        Given a MIntBlock configuration, which looks like ``[{k:1000,ic:...},{k:10000,ic:...}...]``,
        determines the ideal ic time, in nanoseconds
        """
        def isFast(k0):
            fast_k0 = 10_000
            slow_k0 =    100
            if k0 == fast_k0: return True
            if k0 == slow_k0: return False
            else:             return True # is default at firmware side
        
        fast_ic_time =    100_000 # 100 us in ns
        slow_ic_time = 10_000_000 #  10 ms in ns
        areKfast = [ isFast(intConfig.get("k",None)) for intConfig in mIntConfig ]
        return fast_ic_time if all(areKfast) else slow_ic_time
    
    def set_circuit(self,
            carrier_config,
            **further_commands):
        """
        This sets a carrier level configuration. The :mod:`~lucipy.circuits` module and
        in particular the :func:`~lucipy.circuits.Circuit.generate` method can help to
        produce these configuration data structures.
        
        Typically a configuration looks a bit like ``{"/0": {"/U": [....], "/C":[....], ...}}``,
        i.e. the entities for a single cluster. There is only one cluster in LUCIDAC.
        
        :param reset_before: Reset circuit configuration on device before setting the new one.
           Pass ``False`` for incrementally updating the configuration.
        :param sh_kludge: Make a SH Track-Inject cycle (after potential reset) before applying
           the configuration on the LUCIDAC.
        :param calibrate_...: Perform the device calibration scheme. Currently disabled.
        
        .. note::
        
           This also determines the ideal IC time *if* that has not been set
           before (either manually or in a previous run).
        """
        
        from .circuits import Circuit
        if isinstance(carrier_config, Circuit): # or Routing or ...
            carrier_config = carrier_config.generate()
        
        outer_config = dict(
            entity = [self.get_mac()], # str(cluster_index)], # was "0", NOT "/0"
            config =  carrier_config
        )
        
        outer_config = {**outer_config, **self.circuit_options}
        outer_config = {**outer_config, **further_commands}
        
        #print(f"set_circuit {outer_config=}")

        if "/0" in carrier_config:
            cluster_config = carrier_config["/0"]
            if "/M0" in cluster_config and not "ic_time" in self.run_config:
                self.run_config.ic_time = self.determine_idal_ic_time_from_k0s(cluster_config["/M0"])
            
        if "adc_channels" in carrier_config:
            from .simulator import remove_trailing
            num_channels = len(remove_trailing(carrier_config["adc_channels"], None))
            self.set_daq(num_channels=num_channels)
                    
        return self.query("set_circuit", outer_config)
        
    
    def set_config(self, circuit):
        "set_config was renamed to set_circuit in later firmware versions"
        return self.set_circuit(circuit)
    
    def set_circuit_alt(self, circuit):
        # manually decompose /0
        carrier_config = circuit
        if "/0" in carrier_config:
            self.query("set_circuit", {
                "entity": [self.get_mac(), "0"],
                "config": carrier_config["/0"]
            })
            del carrier_config["/0"]
        self.set_circuit(carrier_config)
    
    @staticmethod
    def resolve_path(path, config={}):
        """
        Provides the notational sugar for :meth:`set_by_path`.
        
        The path must point to an entity and not to an element. 
        An entity corresponds to a class on the firmware side which can receive configuration
        "elements". The path can be provided as string, such as ``/0/C``
        or ``0/C`` or as array of strings such as ``["0", "C"]``.
        
        Since this client has no information about the entity structure and the element
        configuration, it cannot accept paths which also include elements, such as
        ``/0/M0/elements/0/ic`` for setting the first integrator value. However, as
        a "nicety" it allows to generate a nested directory structure by introducing
        a "double slash" notation, which is ``/0/M0//elements/0/ic``, or as array,
        ``["0", "M0", "", "elements", "0", "ic"]``. Internally, such a call is then
        translated to the path ``0/M0`` and the configuration dictionary is wrapped as
        ``{ "elements": {"0": {"ic": actual_config } } }``. This way, one can set
        actual elements very conveniently.

        >>> LUCIDAC.resolve_path("/0/M0")
        (['0', 'M0'], {})
        >>> LUCIDAC.resolve_path("foo/bar", "baz")
        (['foo', 'bar'], 'baz')
        >>> LUCIDAC.resolve_path("foo/bar//baz", "kaz")
        (['foo', 'bar'], {'baz': 'kaz'})
        >>> LUCIDAC.resolve_path("foo/bar//bax/bay/baz", {"bir":"bur"})
        (['foo', 'bar'], {'bax': {'bay': {'baz': {'bir': 'bur'}}}})
            
        In the following real world example, all notations are equvialent:
        
        >>> a = LUCIDAC.resolve_path(["0", "M0"], {"elements":{"0": {"ic":0.23, "k":100} } })
        >>> b = LUCIDAC.resolve_path("/0/M0",     {"elements":{"0": {"ic":0.23, "k":100} } })
        >>> c = LUCIDAC.resolve_path("/0/M0//elements/0",         {"ic":0.23, "k":100})
        >>> print(a)
        (['0', 'M0'], {'elements': {'0': {'ic': 0.23, 'k': 100}}})
        >>> assert a == b and b == c

        """
        if isinstance(path, str):
            apath = path.split("/")
            if apath[0] == "": # remove leading slash if it was there
                apath.pop(0)
            if apath.count("") == 1: # entity-element split given
                split_index = apath.index("")
                entity_path, element_path = apath[:split_index], apath[split_index+1:]
                wrapper_config = config
                for e in element_path[::-1]:
                    wrapper_config = { e: wrapper_config }
                return (entity_path, wrapper_config)
            elif apath.count("") > 1:
                raise ValueError(f"'{path}': Invalid entity path given. Entity-Element split '//' can only applied once.")
            path = apath
        return (path, config)
    
    def set_by_path(self, path, config):
        """
        Set element configuration by path.
        
        This is a fine-granular alternative to :meth:`set_circuit`. For the meaning of
        ``path``, see :meth:`resolve_path`.

        Note that the path is always relative to the carrier, not the cluster. That means
        most of the time you want to address the cluster just by ascending with the
        entity path "0".
         
        When providing entities as list, do not prepend entities with a slash, i.e. do
        not write ``[..., "/M0", ...]`` but just  ``[..., "M0", ...]``.
        Slash-prefixed entitiy names happen to take place only in the
        configuration dictionary (second parameter).
        
        All these following lines are equivalent formulations for the same aim:
        
        >>> hc = LUCIDAC("emu:/")
        >>> hc.set_by_path(["0", "M0"], {"elements":{0: {"ic":0.23, "k":100} } })
        >>> hc.set_by_path("/0/M0",     {"elements":{0: {"ic":0.23, "k":100} } })
        >>> hc.set_by_path("/0/M0//elements/0",         {"ic":0.23, "k":100})
        >>> hc.get_circuit()["config"]["/0"]["/M0"]["elements"]["0"]
        {'ic': 0.23, 'k': 100}

        """
        path, config = self.resolve_path(path, config)
        
        cluster_index = 0
        outer_config = {
            "entity": [self.get_mac()] + path, #[self.get_mac(), str(cluster_index)] + path,
            "config": config,
            **self.circuit_options
        }
        return self.query("set_circuit", outer_config)
    
    def set_leds(self, leds_as_integer):
        return self.set_by_path("FP", { "leds": leds_as_integer })
        
    def signal_generator(self, dac=None):
        """
        :args dac: Digital analog converter outputs, as there are two a list with two floats (normalized [-1,+1]) is expected
        """
        return self.set_by_path("FP", {
            "signal_generator":  { "dac_outputs": dac, "sleep": False }
        })
    
    def set_op_time(self, *, ns=0, us=0, ms=0, sec=0, k0fast=0, k0slow=0):
        """
        Sets OP-Time with clear units. Returns computed value, which is just the **sum of
        all arguments**.
        
        Consider the limitations reported in :meth:`start_run`.
        
        Note that this function signature is somewhat comparable to python builtin
        `datetime.timedelta <https://docs.python.org/3/library/datetime.html#timedelta-objects>`_,
        however Python's timedelta has only microseconds resolution which is not as
        highly-resolved as in LUCIDAC.
        
        :param ns: nanoseconds
        :param us: microseconds
        :param ms: milliseconds
        :param sec: seconds
        :param k0fast: units of k0fast, i.e. the fast integrators. Measuring time in
               this units means that within the time ``k0fast=1``` an integrator
               integrates a constant 1 from initial value 0 to final value 1.
        :param k0slow: units of k0slow. A slow integrators computes within time ``k0slow=1``
               a constant 1 from initial vlaue 0 to final value 1.

        """
        optime_ns  = ns + us*1e3 + ms*1e6 + sec*1e9
        optime_ns += k0fast*1e5 + k0slow*1e7
        self.run_config.op_time = int(optime_ns)
        return self.run_config.op_time
    
    #: Sample rates per second accepted by the system.
    allowed_sample_rates = [1, 2, 4, 5, 8, 10, 16, 20, 25, 32, 40, 50, 64, 80, 100, 125, 160, 200, 250, 320, 400, 500, 625, 800, 1000, 1250, 1600, 2000, 2500, 3125, 4000, 5000, 6250, 8000, 10000, 12500, 15625, 20000, 25000, 31250, 40000, 50000, 62500, 100000, 125000, 200000, 250000, 500000, 1000000]
    
    def set_daq(self, *,
            num_channels = None,
            sample_op = None,
            sample_op_end = None,
            sample_rate = None,
            ):
        """
        :param num_channels: Data aquisition specific - number of channels to sample. Between
           0 (no data aquisition) and 8 (all channels)
        :param sample_op: Sample a first point exactly when optime starts
        :param sample_op_end: Sample a last point exactly when optime ends
        :param sample_rate: Number of samples per second. Note that not all numbers are
           supported. A client side check is performed.
        """
        if num_channels != None:
            if not (0 <= num_channels and num_channels <= 8):
                raise ValueError("Require 0 <= num_channels <= 8")
            self.daq_config.num_channels = num_channels

        if sample_op != None: # boolean
            self.daq_config.sample_op = sample_op
        if sample_op_end != None: # boolean
            self.daq_config.sample_op_end = sample_op_end
        
        if sample_rate != None:
            if not sample_rate in self.allowed_sample_rates:
                raise ValueError(f"{sample_rate=} not allowed. Firmware supports only values from the following list: {self.allowed_sample_rates}")
            self.daq_config.sample_rate = sample_rate
       
        return self.daq_config

    def set_run(self, *,
            halt_on_external_trigger = None,
            halt_on_overload = None,
            ic_time = None,
            op_time = None,
            no_streaming = None,
            repetitive = None,
            ):
        """
        :param halt_on_external_trigger: Whether halt the run if external input triggers
        :param halt_on_overload: Whether halt the run if overload occurs during computation
        :param ic_time: Request time to load initial conditions, in nanoseconds.
           This time depends on the k0 factors. However, it is rarely neccessary to tune this
           parameter, once useful values have been used. If not set, a value derived from the
           (first time in object lifetime) configuration set will be used.
        :param op_time: Request time for simulation, in nanoseconds. This is the most important
           option for this method. Note that by a current limitation in the hardware, only
           op_times < 1sec are supported.
        """
        if halt_on_external_trigger != None:
            self.run_config.halt_on_external_trigger = halt_on_external_trigger
        if halt_on_overload != None:
            self.run_config.halt_on_overload = halt_on_overload
        if ic_time != None:
            self.run_config.ic_time = ic_time
        if op_time != None:
            self.run_config.op_time = op_time
        if no_streaming != None:
            self.run_config.no_streaming = no_streaming
        if repetitive != None:
            self.repetitive = repetitive
        return self.run_config

    def start_run(self, clear_queue=True) -> Run:
        """
        Uses the set_run and set_daq as before.
        Returns a Run object which allows to read all data.
        
        :param clear_queue: Clear queue before submitting, making sure any leftover
           repetitive run is wiped.
        """

        start_run_msg = dict(
            id = str(uuid.uuid4()),
            session = None,
            config = self.run_config,
            daq_config = self.daq_config,
            clear_queue = clear_queue
        )
        
        self.slurp() # slurp old run data or similar
        ret = self.query("start_run", start_run_msg)
        if ret:
            raise LocalError(f"Run did not start successfully. Expected answer to 'start_run' but got {ret=}")
    
        return Run(self)
    
    def manual_mode(self, to:str):
        "manual mode control"
        return self.query("manual_mode", dict(to=to))
    
    def master_for(self, *minions):
        """
        Create a master-minion setup with at least one controlled LUCIDAC (minion).
        minions: type LUCIDAC
        """
        return LUCIGroup(self, *minions)

for cmd in LUCIDAC.commands:
    shorthand = (lambda cmd: lambda self, msg={}: self.query(cmd, msg))(cmd)
    shorthand.__doc__ = f'Shorthand for ``query("{cmd}", msg)``, see :meth:`query`.'
    #shorthand = types.MethodType(shorthand, self) # bind function
    if hasattr(functools, "cache") and cmd in LUCIDAC.memoizable:
        shorthand = functools.cache(shorthand)
    if not hasattr(LUCIDAC, cmd):
        setattr(LUCIDAC, cmd, shorthand)

class LUCIGroup:
    """
    Group of LUCIDACs in a master/minion setup. Usage is like
        
    >>> gru    = LUCIDAC("tcp://foo")       # doctest: +SKIP
    >>> kevin  = LUCIDAC("tcp://bar")       # doctest: +SKIP
    >>> bob    = LUCIDAC("tcp://baz")       # doctest: +SKIP
    >>> group  = LUCIGroup(gru, kevin, bob) # doctest: +SKIP
    >>> group.set_circuit(...)              # doctest: +SKIP
    >>> group.start_run() ...               # doctest: +SKIP
    """
    
    def __init__(self, master: LUCIDAC, *minions: LUCIDAC):
        self.master = master
        self.minions = minions
        for minion in self.minions:
            res = minion.manual_mode("minion")
            if res:
                raise ValueError(f"Failed to make {minion} a minion, returned {res}")
    
    def __getattr__(self, attr):
        if attr not in self.__dict__:
            # for the moment, we forward every command only to the master.
            # Later, we want to collect results, such as from "get_entities".
            # However, in the moment this does not even work on free floating teensies with
            # LUCIDAC REV1 hardware, cf. hybrid-controller#145.
            return getattr(self.master, attr)
        
    
    

if __name__ == "__main__":
    # simple example demonstrator
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-e", "--endpoint", help="Lucidac endpoint URL (such as tcp:/192.168.1.1:5732 or serial:/dev/ttyACM0)")
    args= parser.parse_args()
    hc = LUCIDAC(args.endpoint)
    
    import IPython
    IPython.embed()
    
# Other simple example:
#    hc = HybridController("192.168.68.60", 5732)
#    status = hc.query('status')
#    print(status)



#
# This kind of tests should go against the emulator the same way as against real hardware.#
# Well now at least this file goes against real hardware
#
# Running tests in this file requires the environment variable LUCIDAC_ENDPOINT
# set to a suitable value, for instance "tcp://192.168.150.229:5732"
#

import pytest
import os

from lucipy import LUCIDAC, Circuit, Simulation
from lucipy.simulator import remove_trailing

from fixture_circuits import circuit_sinus, measure_ramp, measure_ics

from itertools import product
import numpy as np

def requires_hardware(f):
    return pytest.mark.skipif(
        not "LUCIDAC_ENDPOINT" in os.environ,
        reason=f"Hardware tests require env var LUCIDAC_ENDPOINT"
    )(f)

@pytest.fixture
def hc():
    hc = LUCIDAC()
    hc.reset_circuit()
    yield hc
    hc.sock.sock.close() # or similar
    
@requires_hardware
def test_empty_configuration(hc):
    hc.reset_circuit()
    conf = Circuit().generate()
    print(conf)
    hc.set_circuit(conf) # this shall not raise, otherwise the Circuit generation is broken

# tests the protocol and configuration readout
@requires_hardware
def test_set_circuit_for_cluster(hc):
    set_conf_cluster = circuit_sinus().generate(skip="/M1") # Test Hardware has no M1!)
    hc.set_circuit(set_conf_cluster)
    get_conf_cluster = hc.get_circuit()["config"]
           
    # canonicalize I block defaults:
    for i,v in enumerate(get_conf_cluster["/0"]["/I"]["outputs"]):
        if not v:
            get_conf_cluster["/0"]["/I"]["outputs"][i] = []
            
    # for the moment, get rid of U constant output. TODO Needs treatment!
    del get_conf_cluster["/0"]["/U"]["constant"]
    
    # get rid of M1 block
    del get_conf_cluster["/0"]["/M1"]
    
    # get rid of SH block
    if "/SH" in get_conf_cluster["/0"]:
        del get_conf_cluster["/0"]["/SH"]
            
    print(f"{set_conf_cluster['/0']=}")
    print(f"{get_conf_cluster['/0']=}")
    
    ## Differences are still in the U-block. Look carefully.
    ## Probably we have an old an unsuitable commit.
    
    assert set_conf_cluster["/0"] == get_conf_cluster["/0"]

# tests the protocol and configuration readout
@requires_hardware
def test_set_adc_channels(hc):
    c = Circuit()
    c.set_adc_channels([0,1,2])
    set_conf = c.generate(skip="/M1")  # Test Hardware has no M1!
    print(f"{set_conf=}")
    hc.set_config(set_conf)
    get_conf = hc.get_circuit()["config"]
    
    # canonicalize: Remove trailing "None"
    get_adc_channels = remove_trailing(get_conf["adc_channels"])    
    
    print(f"{get_conf=}")
    print(f"{get_adc_channels=}")
    assert get_adc_channels == c.adc_channels

@requires_hardware
@pytest.mark.parametrize("ic,slow", product([-1, -0.5, 0, +0.5, +1], [False, True]))
def test_ics(hc, ic, slow):
    hc.reset_circuit()
    hc.manual_mode("ic")
    measure_ics(hc, ic, slow, do_assert=True)

@requires_hardware
@pytest.mark.parametrize("slope,lane", product([+1,-1], range(32)))
def SKIP_test_ramp(hc, slope, lane):
    valid_endpoint, valid_evolution == measure_ramp(slope, lane) # const_value=-1
    assert valid_endpoint
    assert valid_evolution



# TESTs to add:
#
#
#  1) Very short runtime: daq
#  2) Very long runtime: daq



#!/usr/bin/env python3

"""
This is a small interactive command line script for discovering/detecting LUCIDACs.
It will look for the Teensy Microcontroller (Hybrid Controller) both directly
connected over USB Serial Device as well as Network services announced by
MDNS/Zeroconf over the network (local IPv4 broadcast domain). The network lookup
happens typically within a few hundred milliseconds. If you want to wait for more
then one device, use the --all option.
"""

# all python included
import asyncio, logging, socket, sys, argparse, inspect, ast, pathlib, time, collections, itertools, urllib, re
from typing import Any, Optional, List, cast, Iterator

try:
    # pip install pyserial
    import serial, serial.tools.list_ports
except ModuleNotFoundError:
    serial = None

try:
    # pip install zeroconf
    # An OS-independent all-python zeroconf/bonjour client
    from zeroconf import IPVersion, ServiceStateChange, Zeroconf
    from zeroconf.asyncio import (
        AsyncServiceBrowser,
        AsyncServiceInfo,
        AsyncZeroconf,
        AsyncZeroconfServiceTypes,
    )
except ModuleNotFoundError:
    Zeroconf = None

verbosity = 0
err = lambda msg: print(msg, file=sys.stderr)
log = lambda level, msg: print(msg, file=sys.stderr) if verbosity >= level else None
v   = lambda msg: log(1, msg)
vv  = lambda msg: log(2, msg)


class Endpoint:
    """
    This class models the LUCIDAC/REDAC endpoint URI convention in lucipy.
    Given there is no proper URI object in python, we make use of
    ``urllib.urlparse``. This class is used for instance in
    :func:`lucipy.synchc.endpoint2socket`.

    Create instances either directly or via helpers which prepend
    the correct scheme:

    >>> Endpoint("tcp://123.123.123.123:7890")
    Endpoint("tcp://123.123.123.123:7890")
    >>> Endpoint.fromJSONL("localhost", 1234)
    Endpoint("tcp://localhost:1234")

    For serial devices, there is a little extra work done ontop of
    URL parsing, which makes the syntax less weird in contexts where
    URIs are typically not used. Serial devices just don't follow the
    double slash convention but more a kio/gfs single slash
    convention:

    >>> Endpoint("serial:/dev/ttyACM0") # device file at Linux/Mac
    Endpoint("serial:/dev/ttyACM0")
    >>> Endpoint.fromDevice("/dev/ttyACM0")
    Endpoint("serial:/dev/ttyACM0")
    >>> Endpoint("serial:/COM0")
    Endpoint("serial:/COM0")
    >>> Endpoint.fromDevice("COM0") # port name at Windows
    Endpoint("serial:/COM0")

    Endpoints with scheme only are fine and used in the code. Note that
    schemes are the only part of the URL which is always canonically
    lowercased:

    >>> Endpoint("EMU:")
    Endpoint("emu:")

    An example of an endpoint which uses all fields:

    >>> e = Endpoint("TCP://myuser:mypass@FOO.bar:4711?flitz=bums&baz=bla")
    >>> e.user
    'myuser'
    >>> e.args
    {'flitz': 'bums', 'baz': 'bla'}
    >>> e
    Endpoint("tcp://myuser:foo.bar:4711?flitz=bums&baz=bla")
    """
    default_tcp_port = 5732
    
    def __init__(self, endpoint):
        if isinstance(endpoint, Endpoint):
            endpoint = endpoint.url()
        if not isinstance(endpoint, str):
            raise ValueError("Expecting endpoint URL as string")
            
        result = urllib.parse.urlparse(endpoint)
        #result = re.match("(?P<scheme>[a-zA-Z0-9]+):(?P<sep>/?/?)(?P<userpass>[^@:]+(:[^@]+)?)...

        #: Scheme, such as "serial", "tcp", etc.
        self.scheme = result.scheme
        #: Username for login, if present. None is a valid value.
        self.user = result.username
        #: Password for login, if present. None is a valid value.
        self.password = result.password
        #: Host or device name. Note that hostnames are transfered to lowercase while
        #: pathnames will not.
        self.host = (result.hostname or "") + (result.path or "")
        #: TCP/IP Port as integer. If not given, defaults to default TCP port.
        self.port = int(result.port or self.default_tcp_port)
        #: Further query arguments from the URL
        self.args = urllib.parse.parse_qs(result.query, keep_blank_values=True)
        # improve for queries such as ?foo -> should result in foo=True
        #                             ?foo=bar -> should not result in foo=["bar"] but foo=bar
        for k in self.args.keys():
            if len(self.args[k]) == 0:
                self.args[k] = True
            if len(self.args[k]) == 1:
                self.args[k] = self.args[k][0] # unwrap

        # fixes for serial scheme, mainly neccessary because
        # result.hostname is lowercased, which is bad for "ttyACM0" or "COM0".
        # Will also fix problems such as serial://dev/foo resulting in host= "dev/foo"
        posix = re.match("serial:/?(/.+)", endpoint, re.IGNORECASE)
        win = re.match("serial:/?/(.+)", endpoint, re.IGNORECASE)
        if posix:
            self.host = posix.group(1)
        elif win:
            self.host = win.group(1)

        #if self.scheme == "serial" and "/" in self.host and self.host[0] != "/":
            # fix Unix absolute filenames: serial://dev/foo would result in host = "dev/foo"
            # however serial:/dev/foo gets correct "/dev/foo". This corrects the double slash version.
        #    self.host = "/" + self.host
        
        #if not self.scheme:
        #    raise ValueError(f"Invalid Endpoint '{endpoint}' is not a string-encoded URL.")

    @staticmethod
    def fromDevice(device_name):
        "Initialize for a device name"
        sep_host = "//"+device_name if device_name[0] != "/" else device_name
        return Endpoint(f"serial:{sep_host}")
    
    @staticmethod
    def fromJSONL(addr, port=None):
        "Initialize for a TCP addr/port address tuple"
        return Endpoint(f"tcp://{addr}" + (f":{port}" if port else ""))
    
    def url(self):
        s = self.scheme + ":"
        if self.host and self.host[0] != "/":
            s += "//"
        if self.user and not self.password:
            s += self.user + ":" + self.password + "@"
        elif self.user:
            s += self.user + ":"
        s += self.host
        # the way attaching the port to the host won't work for websockets
        # (think of Endpoint("http://foo.bar:80/bla.ws")) but we don't support them anyway.
        if self.port != self.default_tcp_port:
            s += ":" + str(self.port)
        if self.args:
            s += "?" + "&".join([f"{k}={v}" for k,v in self.args.items()])
        return s
    
    def __repr__(self):
        return 'Endpoint("' + self.url() +  '")'
    
    def __str__(self):
        return self.__repr__()

def can_resolve(hostname, target_ip:str=None) -> Optional[str]:
    "gethostbyname but without the exceptions, i.e. with a boolean representation of the return value"
    try:
        return socket.gethostbyname(hostname)
    except socket.error:
        return None

def can_resolve_to(hostname, expected_ip:str):
    "Checks whether the host system can resolve a given (zeroconf) DNS name"
    return can_resolve(hostname) == expected_ip

class ZeroconfDetector:
    def __init__(self, timeout_ms=500):
        # self.search_for = "lucidac-AA-BB-CC" # something which results in an abortion condition!
        if not Zeroconf:
            raise ModuleNotFoundError("Constructing a ZeroconfDetector object requires zeroconf, install with 'pip install zeroconf'")
        self.aiobrowser: Optional[AsyncServiceBrowser] = None
        self.aiozc: Optional[AsyncZeroconf] = None
        self.results: List[Endpoint] = []
        self.timeout_ns = timeout_ms*1000
        
    def on_service_state_change(self, zeroconf, service_type: str, name: str, state_change) -> None:
        # types are actually
        # zeroconf: Zeroconf, service_type: str, name: str, state_change: ServiceStateChange
        # but not using them for avoiding dependencies.
        vv(f"Service {name} of type {service_type} state changed: {state_change}")
        if state_change is not ServiceStateChange.Added:
            return
        asyncio.ensure_future(self._enqueue_service_info(zeroconf, service_type, name))
        
    async def _enqueue_service_info(self, zeroconf, service_type: str, name: str):
        info = AsyncServiceInfo(service_type, name)
        await info.async_request(zeroconf, 3000)
        vv("Zeroconf found: %r" % (info))
        if info:
            for addr in info.parsed_scoped_addresses():
                v(f"Found {info.server} resolving to {addr}; TCP Port {info.port}")
                # Use Hostname only if the system can resolve it (functional mDNS setup).
                # Most Windows, Mac OS and Linux systems can do it, some (in particular Linux) cannot.
                addr = info.server if can_resolve_to(info.server, expected_ip=addr) else addr
                endpoint = Endpoint.fromJSONL(addr, info.port)
                if self.timeout_ns == 0:
                    print(endpoint)
                else:
                    self.results.append(endpoint)
                    vv(self.results)

    async def start(self) -> None:
        "Starts Zeroconf browser detection. Returns after timeout."
        self.aiozc = aiozc = AsyncZeroconf(ip_version=IPVersion.V4Only)
    
        services = [ "_lucijsonl._tcp.local." ] # not even _http
        v("...Scanning local broadcast network for network, press Ctrl-C to exit..." % services)

        self.aiobrowser = AsyncServiceBrowser(
            self.aiozc.zeroconf, services, handlers=[self.on_service_state_change]
        )
        
        started = time.time_ns()
        while 0 == self.timeout_ns or (time.time_ns() - started) < self.timeout_ns: # or len(self.results) < 1:
            print("waiting...")
            await asyncio.sleep(.5)
        
        return self.results
        
    async def stop(self) -> None:
        "Stops Zeroconf browser detection"
        assert self.aiozc is not None
        assert self.aiobrowser is not None
        await self.aiobrowser.async_cancel()
        await self.aiozc.async_close()
          
    def sync_start(self):
        "merely a helper to trigger and endless run"
        
        #loop = asyncio.get_event_loop()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            return loop.run_until_complete(self.start())
        except KeyboardInterrupt:
            loop.run_until_complete(self.stop())
            return self.results


def detect_usb_teensys() -> List[Endpoint]:
    "Yields all found endpoints on local system using serial.tools.list_ports, requires pyserial"
    teensy_vid = 0x16C0
    teensy_pid = 0x0483
    
    if not serial:
        raise ModuleNotFoundError("lucipy.detect.detect_usb_teensys for USB requires pyserial, install with 'pip install pyserial'")
    
    found = []
    for port in serial.tools.list_ports.comports(): 
        if port.pid == teensy_pid and port.vid == teensy_vid:
            v(f"Serial device at {port.device} - {port.hwid}") # sth like "USB VID:PID=16C0:0483 SER=15240110 LOCATION=1-3:1.0"
            found.append(Endpoint.fromDevice(port.device)) # sth like "/dev/ttyACM0" at Linux/Mac
            
    # TODO: Has to connect and make sure it is a Teensy belonging to a LUCIDAC
    #       and not to something else. i.e. make sure it speaks the JSONL protocol.
    
    return found

def detect_network_teensys(zeroconf_timeout=500) -> List[Endpoint]:
    "Yields all endpoints in the local broadcast domain using Zeroconf, requires python zeroconf package"
    if not Zeroconf:
        raise ModuleNotFoundError("lucipy.detect.detect_network_teensys requires zeroconf, install with 'pip install zeroconf'")
    
    return ZeroconfDetector(zeroconf_timeout).sync_start()

def detect(single=False, prefer_network=True, zeroconf_timeout=500):# -> Optional[Endpoint | List[Endpoint]]:
    """
    Yields or returns possible endpoints using all methods. This function will raise an ModuleNotFoundError
    if a library is not available which might have found more.

    :param single: Return only first found instance or None, if nothing found. If this
         option is False, this function will return an array of endpoints discovered using all methods.
    :param zeroconf_timeout: Maximum search time: How long to wait for zeroconf answers,
         in milliseconds. Set to 0 or None for unlimited search.
    :param prefer_network: Return network result first. Typically a TCP/IP connection is
         faster and more reliable then the USBSerial connection.
    """
    res = []
    singlize = lambda res: (res[0] if len(res) else None) if single else res
    if prefer_network:
        res += detect_network_teensys(zeroconf_timeout)
    if single and len(res):
        return singlize(res)
    if True:
        res += detect_usb_teensys()
    if single and len(res):
        return singlize(res)
    if not prefer_network:
        res += detect_network_teensys(zeroconf_timeout)
    return singlize(res)

if __name__ == '__main__':
    #logging.basicConfig(level=logging.INFO)
    file_docstring = ast.get_docstring(ast.parse(pathlib.Path(__file__).read_text()), clean=True)    
    parser = argparse.ArgumentParser(description="Scanner to discover locally or network-attached LUCIDAC analog/digital hybrid computers.",
        epilog = file_docstring,
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-v', '--verbose', action='count', default=0, help="Verbosity, add more -v for more verbose output")
    parser.add_argument('-a', '--all', action='store_true', help='Scan unlimited, try to find more then one device in the network (stop with CTRL-C)')
    args = parser.parse_args()
    verbosity = args.verbose
    for res in detect(zeroconf_timeout=0 if args.all else 1000):
        print(res)



import pytest
from lucipy import LUCIDAC, Circuit, Route, Connection, Simulation, Emulation
from fixture_circuits import circuit_constant2acl_out

def test_constant_detection_in_simulation():
    const = Circuit()
    const.use_constant()
    # constant from clane=14 -> DO end up in 16<=lanes<32
    # constant from clane=15 -> DO end up in  0<=lanes<16
    const.add( Route(14, 0, 0.1, 0) ) # must not work
    const.add( Route(14,16, 0.2, 1) ) # must work
    const.add( Route(15,17, 0.3, 2) ) # must not work
    const.add( Route(15, 2, 0.4, 3) ) # must work
    sim = Simulation(const)
    print(sim.constant)
    assert all(sim.constant[0:4] == [ 0., 0.2, 0., 0.4])
    
def do_NOT_YET_test_measure_acl_constant():
    coeff0 = -0.5
    coeff1 = +0.5
    
    sim = Simulation(circuit_constant2acl_out(coeff0, coeff1))
    irrelevant_state = [0.]*8
    acl_outs = sim.acl_out_values(irrelevant_state)
    assert acl_outs[0] == coeff0
    assert acl_outs[1] != coeff1

def test_integrator_chain():
    c = Circuit()
    i = c.ints(7)
    cnst = c.const()
    c.connect(cnst, i[0], weight=-1)
    for k in range(6):
        c.connect(i[k], i[k+1], weight=-(k+2))
    res = Simulation(c).solve_ivp(1, dense_output=True)
    
    import numpy as np
    tfine = np.linspace(0, 1)
    for k in range(7):
        # final one was reached
        assert np.isclose(res.y[k][-1], 1, atol=1e-3)
        # on coarse support points. When plotted with option "o-"
        # this looks weird because of the few support points.
        assert np.allclose(res.t**(k+1), res.y[k], atol=1e-3)
        # on a finer one which looks, when plotted, more like the solution
        assert np.allclose(tfine**(k+1), res.sol(tfine)[k], atol=1e-3)
        # (note, bigger tolerance is needed for bigger k)

def test_multipliers():
    c = Circuit()
    it = c.int() # integrates a "time"
    m1, m2, m3 = c.muls(3)
    cnst = c.const()
    
    a, b = 0.3, -0.25
    ab = a*b
    
    c.connect(cnst, it, weight=-1)
    
    # m1 = const * const
    c.connect(cnst, m1.a, weight=a)
    c.connect(cnst, m1.b, weight=b)
    
    # m2 = t * const
    c.connect(cnst, m2.a, weight=a)
    c.connect(it, m2.b)
    
    # m3 = t * t
    c.connect(it, m3.a)
    c.connect(it, m3.b)
    
    sim = Simulation(c)
    res = sim.solve_ivp(1, dense_output=True)
    # res.y holds integrator values in shape (num_integrators, num_timesteps)
    
    import numpy as np
    resm = np.array([sim.Mul_out(yt) for yt in res.y.T]).T
    # resm is the same shape (num_multipliers, num_timesteps)
    
    assert np.allclose(res.t, res.y[0]), "time ramp not properly integrated"
    assert np.allclose(resm[0], ab) # m1
    assert np.allclose(resm[1], a * res.t) # m2
    assert np.allclose(resm[2], res.t * res.t) # m3
    
    # can do the same also on "more continous" time
    tfine = np.linspace(0,1)
    resfine = res.sol(tfine)
    resmfine = np.array([sim.Mul_out(ytf) for ytf in resfine.T]).T
    
    assert np.allclose(tfine, resfine[0])
    assert np.allclose(resmfine[0], ab) # m1
    assert np.allclose(resmfine[1], a * tfine) # m2
    assert np.allclose(resmfine[2], tfine * tfine) # m3

@pytest.mark.parametrize("slow", [False, True])
def test_ramp(slow):
    # This circuit uses the constant giver for integrating over a constant
    # if slow, will use slow integrator.
    
    slow_factor = 100 if slow else 1 # difference between k0fast and k0slow
    
    ic = +1
    slope = -1
    t_final = 2 * slow_factor
    expected_result = -t_final * slope/slow_factor - ic
    assert expected_result == +1

    ramp = Circuit()
    i = ramp.int()
    assert i.out == 0
    
    c = ramp.const()
    assert c.out == 14
    
    ramp.connect(c, i, weight=slope)
    ramp.set_ic(0, ic)
    ramp.set_k0_slow(0, slow)
    
    sim = Simulation(ramp)
    assert sim.constant[0] == slope
    assert all(sim.constant[1:] == 0)
    
    res = sim.solve_ivp(t_final)
    
    assert res.y[0, 0] == -ic
    
    import numpy as np
    assert np.isclose(res.y[0,-1], expected_result)



#!/usr/bin/env python3

"""
Lucipy Circuits: A shim over the routes.

This class structure provides a minimal level of user-friendlyness to allow
route-based programming on the LUCIDAC. This effectively means a paradigm
where one connects elements from math blocks to each other, with a coefficient
inbetween. And the assignment throught the UCI matrix is done greedily by
"first come, first serve" without any constraints.

The code is heavily inspired by the LUCIGUI lucisim typescript compiler,
which is however much bigger and creates an AST before mapping.

In contrast, the approach demonstrated here is not even enough for REV0 connecting
ExtIn/ADC/etc. But it makes it very simple and transparent to work with routes
and setup the circuit configuration low level.
"""

import functools, operator, textwrap, pprint, itertools, random
from collections import namedtuple
from typing import get_args, List, Dict, Union, Optional

# like sum(lst, []) but accepts generators instead of lists
flatten = lambda lst: functools.reduce(operator.iconcat, lst, [])
find = lambda crit, default, lst: next((x for x in lst if crit(x)), default)
clean = lambda itm: [ k[0] if len(k)==1 else (None if len(k)==0 else k) for k in itm ]
notNone = lambda thing: thing != None

def window(seq, n=2):
    "Returns a sliding window (of width n) over data from the iterable"
    "   s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...                   "
    it = iter(seq)
    result = tuple(itertools.islice(it, n))
    if len(result) == n:
        yield result
    for elem in it:
        result = result[1:] + (elem,)
        yield result


def next_free(occupied: List[bool], criterion=None, append_to:int=None) -> Optional[int]:
    """
    Looks for the first False value within a list of truth values.

    >>> next_free([1,1,0,1,0,0]) # using ints instead of booleans for brevety
    2

    If no more value is free in list, it can append up to a given value

    >>> next_free([True]*4, append_to=3) # None, nothing free
    >>> next_free([True]*4, append_to=6)
    4

    :arg criterion: Callback which gets the potential value and can decide whether this
      value is fine. Can be used for constraining an acceptable next free lane.
    """
    for idx, val in enumerate(occupied):
        if not val and (criterion(idx) if criterion else True):
            return idx
    return len(occupied) if append_to != None and len(occupied) < append_to else None

## Note on naming:
## "in" is a reserved keyword in python, so all inputs are called a,b,c,... even if there is ony one

Int = namedtuple("Int", ["id", "out", "a"])          #: Integrator, integrates input "a", output to "out"
Mul = namedtuple("Mul", ["id", "out", "a", "b"])     #: Multiplier, multiplies inputs "a" with "b", output to "out"
Id  = namedtuple("Id",  ["id", "out", "a"])          #: Identity element, just passes input "a" to output "out"
Const = namedtuple("Const", ["id", "out" ])          #: Constant giver. output on (cross lane "out")
Front = namedtuple("Out", ["id", "lane"])            #: Front panel input/output (ACL_IN/ACL_OUT) on lane "lane"
Ele = Union[Int,Mul,Const,Id,Front]                  #: type of any kind of element defined so far
isEle = lambda thing: isinstance(thing, get_args(Ele))


class DefaultLUCIDAC:
    """
    This describes a default LUCIDAC REV1 setup with
    
    - M0 = MIntBlock (8 integrators)
    - M1 = MMulBlock (4 multipliers and ID-lanes)
    
    In particular, note that the clanes of M0 and M1 have switched during the
    transition from REV0 to REV1 hardware.
    """
    
    num_int = 8
    num_mul = 4
    num_const = 2 # REV1 has constant givers in U block, this is purely "virtual" here
    num_id  = 4 # REV1 identity elements in MMul block
    num_acls = 8 # REV1 ACL_IN and ACL_OUT ports (each)
    
    MIntOffset = 0       # M0 block
    MMulOffset = num_int # M1 block
    acl_offset = 24 # where the ACL lanes start

    @staticmethod
    def reservoir(default_value=False):
        return {
            Int:   [default_value]*DefaultLUCIDAC.num_int,
            Mul:   [default_value]*DefaultLUCIDAC.num_mul,
            Const: [default_value]*DefaultLUCIDAC.num_const,
            Front: [default_value]*DefaultLUCIDAC.num_acls,
            Id:    [default_value]*DefaultLUCIDAC.num_id,
        }
    
    @classmethod
    def make(cls, t:Ele, idx):
        """
        A factory for the actual elements.
    
        >>> DefaultLUCIDAC().make(Int, 1)
        Int(id=1, out=1, a=1)
        >>> DefaultLUCIDAC().make(Mul, 3)
        Mul(id=3, out=11, a=14, b=15)
        """
        if t == Int:
            return Int(idx, cls.MIntOffset+idx, cls.MIntOffset+idx)
        if t == Mul:
            return Mul(idx, 
                cls.MMulOffset + idx,
                cls.MMulOffset + 2*idx,
                cls.MMulOffset + 2*idx+1)
        if t == Id:
            # identity elements on Mul-Block map the first 4 MMul inputs on the last 4 MMul outputs
            return Id(idx, cls.MMulOffset+4+idx, cls.MMulOffset+idx)
        if t == Const:
            # Convention for REV1 constants:
            # const(idx=0) => taken from clane 14 => has to be used in lanes  0..15
            # conts(idx=1) => taken from clane 15 => has to be used in lanes 16..31
            return Const(idx, 14 + idx)
        if t == Front:
            return Front(idx, cls.acl_offset + idx)
        raise ValueError(f"Illegal value for {t=}. Need a type like Int/Mul/Id/Const/Front from lucipy.circuits")
        
    @staticmethod
    def populated():
        "An unsorted list of all allocatable computing elements"
        return flatten([ [ DefaultLUCIDAC.make(t,i) for i,_ in enumerate(v) ] for t,v in DefaultLUCIDAC.reservoir().items() ])
    
    
    @classmethod
    def resolve_mout(cls, idx, reservoir):
        """
        Simplistic way to map mul block outputs to a reservoir.
        This is sensitive on M0 and M1 block positions.
        """
        if idx < 0:  raise ValueError(f"0 < {idx=} < 16 too small")
        if idx < 8:  return reservoir[Int][idx]
        if idx < 12: return reservoir[Mul][idx - cls.MMulOffset]
        if idx < 16: return reservoir[Id][idx - 12]
        else:        raise ValueError(f"0 < {idx=} < 16 too large")

class Reservoir:
    """
    This is basically the entities list, tracking which one is already
    handed out ("allocated") or not.

    Note that the Mul/Int classes only hold some integers. In contrast, the
    configurable properties of the stateful computing element (Integrator)
    is managed by the MIntBlock class below.
    """
    allocated: Dict[Ele,List[bool]]
    
    def __init__(self, allocation=None, **kwargs):
        super().__init__(**kwargs)  # forwards all unused arguments
        # an idea was to look that up as in luci.get_entities(), however now we keep it simple
        self.allocated = DefaultLUCIDAC.reservoir() if not allocation else allocation


    def alloc(self, t:Ele, id=None):
        """
        Allocate computing elements.
        
        >>> r = Reservoir()
        >>> r.alloc(Int,1)
        Int(id=1, out=1, a=1)
        >>> r.alloc(Int)
        Int(id=0, out=0, a=0)
        >>> r.alloc(Int)
        Int(id=2, out=2, a=2)
        """
        try:
            lst = self.allocated[t]
            idx = id if id != None else next_free(lst)
            if idx == None:
                raise ValueError(
                    f"No more free Computing Elements for type {t}, all {len(lst)} occupied!"
                    if id != None else
                    f"Compute Element {t} number {id} is already allocated")
            self.allocated[t][idx] = True
            return DefaultLUCIDAC.make(t, idx)
        except KeyError:
            raise TypeError(f"Computing Element Type {t} not supported. Valid ones are {', '.join(map(str, self.allocated.keys()))}")
        except IndexError:
            raise ValueError(f"Have only {len(self.allocated[t])} Computing Elements of Type {t} available, inexistent id {id} requested.")

    # TODO: Rename to "integrator" in order to make sure
    #       it is not misunderstood as "integration"
    def int(self, id=None):
        "Allocate an Integrator. If you pass an id, allocate that particular integrator."
        return self.alloc(Int, id)

    # TODO: Rename to "multiplier" in order to make sur
    #       it is not misunderstood as "multiplication"
    def mul(self, id=None):
        "Allocate a Multiplier. If you pass an id, allocate that particular multiplier."
        return self.alloc(Mul, id)
    
    def const(self, id=None):
        "Allocates a constant. Also turns on the constant giver scheme in general."
        if hasattr(self, "use_constant"):
            self.use_constant() # use_constant is a method of the Routing and Circuit classes
        return self.alloc(Const, id)
    
    def identity(self, id=None):
        "Allocates one Identity element"
        return self.alloc(Id, id)
    
    # some more fun
    def ints(self, count):
        "Allocate count many integrators"
        return [self.int() for x in range(count)]

    def muls(self, count):
        "Allocate count many multipliers"
        return [self.mul() for x in range(count)]
    
    def identities(self, count):
        "Allocate count many identifiers"
        return [self.identity() for x in range(count)]
    
    def front_panel(self, id=None):
        "Allocates an ACL_IN/ACL_OUT Front panel input/output"
        # Do not do the following as we don't know if the output or input is requested!
        #if hasattr(self, "set_front_panel"):
        #    self.set_front_panel(id, as_input=True) # method in Routing and Circuit classes
        return self.alloc(Front, id)

class Route:
    """
    Routes are the essential building block of this circuit API.
    A list of routes is a way to describe the sparse system matrix (UCI-matrix).
    Sparse matrix values ``A[i,j]`` can be described with lists of 3-tuples
    ``(i,j,A)``, i.e. the position and value. This is what ``(uin,iout,coeff)``
    basically describe. Additionally, there is the ``lane`` which describes the
    internal structure of the UCI matrix. The maximum of 32 lanes corresponds to
    the fact that only 32 elements within the system matrix can be nonzero.

    Python allows for any types for the tuple values. Regularly, instances of
    the computing elements (:class:Int, :class:Mul, etc) are used at the
    ``uin`` and ``iout`` slots.

    This compiler knows the concept of "not yet placed" routes. This are routes
    where ``lane == None``. Some of our codes refer to such routes as "logical"
    in contrast to "physically" placed routes. In this code, unplaced routes are
    called "Connection", i.e. they can be generated with the :func:`Connection`
    function.
    """
    
    #: iout constant in order to not connect.
    do_not_connect = -1
    
    def __init__(self, uin, lane, coeff, iout):
        self.uin = uin
        self.lane = lane
        self.coeff = coeff
        self.iout = iout
    
    def __repr__(self):
        return f"Route(uin={self.uin}, lane={self.lane}, coeff={self.coeff}, iout={self.iout})"

    def __iter__(self):
        yield self.uin
        yield self.lane
        yield self.coeff
        yield self.iout
        
    def __eq__(self, other):
        return list(self) == list(other)

    def resolve(self):
        """
        Resolve elements at uin/iout to numbers. This only works if they have the attributes a/b/out.
        For more fancy elements, it is subject of the caller to resolve them properly.
        """
        if isEle(self.uin):
                self.uin = self.uin.out

        if isEle(self.iout):
            if hasattr(self.iout, "b"):
                # element with more then one input
                raise ValueError(f"Please provide input port for {self.iout=} in {self}")
            elif hasattr(self.iout, "a"):
                self.iout = self.iout.a
            else:
                raise ValueError(f"Element has no inputs. Probably mixed up sinks and sources?")
    
    def sanity_list(self):
        errors = []
        if None in self:
            errors.append(f"Route contains None values.")
        if not self.uin in range(0,16) and self.uin != self.do_not_connect:
            errors.append(f"Uin out of range in {self}")
        if not self.lane in range(0,32):
            errors.append(f"Lane out of range in {self}")
        if not (-10 <= self.coeff and self.coeff <= +10):
            errors.append(f"Coefficient out of range in {self}")
        if not self.iout in range(0,32) and self.iout != self.do_not_connect:
            errors.append(f"Iout out of range in {self}")
        return errors
            
    def sanity_raise(self):
        error = ",".join(self.sanity_list())
        if error:
            raise ValueError(error)
        


def Connection(source:Union[Ele,int], target:Union[Ele,int], weight=1):
    """
    Syntactic sugar for a "logical route", i.e. a Route without a lane.        

    >>> r = Reservoir()
    >>> I1, M1 = r.int(), r.mul()
    >>> Connection(M1.a, I1.out)
    Route(uin=8, lane=None, coeff=1, iout=0)
    >>> Connection(M1, I1)
    Route(uin=Mul(id=0, out=8, a=8, b=9), lane=None, coeff=1, iout=Int(id=0, out=0, a=0))

    """
    return Route(source, None, weight, target)

class MIntBlock:
    """
    Stateful configuration about all the MIntBlock.
    """
    #ics : List[float]
    #k0s : List[int]
    
    slow =    100
    fast = 10_000
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)  # forwards all unused arguments
        self.ics = [0.0]*DefaultLUCIDAC.num_int
        self.k0s = [self.fast]*DefaultLUCIDAC.num_int
        
    def randomize(self):
        for i in range(DefaultLUCIDAC.num_int):
            self.ics[i] = random.uniform(-1.0, +1.0)
            self.k0s[i] = random.choice([self.slow, self.fast])

    def set_ic(self, el:Union[Int,int], val:float):
        if val < -1 or val > +1:
            raise ValueError(f"Integrator ICs can only be in range [-1,+1], {val} given for element {el}. Consider rescaling your ODE or coefficient upscaling.")
        el = el.id if isinstance(el, Int) else el
        self.ics[el] = val
    
    def set_k0(self, el:Union[Int,int], val:float):
        el = el.id if isinstance(el, Int) else el
        self.k0s[el] = val
    
    def set_k0_slow(self, el:Union[Int,int], val:bool):
        self.set_k0(el, self.slow if val else self.fast)
        
    def generate(self):
        "Returns configuration for the MIntBlock"
        return {
            "elements": [dict(k=k, ic=ic) for k,ic in zip(self.k0s, self.ics)]
        }

    def load(self, config):
        """
        Inverts what :meth:`generate` is doing.
    
        >>> b = MIntBlock()
        >>> b.randomize()
        >>> config = b.generate()
        >>> c = MIntBlock()
        >>> c.load(config) == c # chainable
        True
        >>> assert b.ics == c.ics
        >>> assert b.k0s == c.k0s
        """
        for idx, integrator in enumerate(config["elements"]):
            if "k" in integrator:
                self.k0s[idx] = integrator["k"]
            if "ic" in integrator:
                self.ics[idx] = integrator["ic"]
        return self
    
    def to_pybrid_cli(self):
        """
        Generate the Pybrid-CLI commands as string out of this Route representation
        """
        ret = []
        ret += [f"set-element-config carrier/0/M0/{i} ic {val}" for i,val in enumerate(self.ics) if val != 0]
        ret += [f"set-element-config carrier/0/M0/{i} k {val}" for i,val in enumerate(self.ics) if self.k0s != self.fast]
        return "\n".join(ret)

class Probes:
    """
    Models the LUCIDAC Carrier ADC channels and Front Panel inputs (ACL_select).
    """
    
    def __init__(self, acl_select=None, adc_channels=None, **kwargs):
        super().__init__(**kwargs)  # forwards all unused arguments
        self.acl_select = acl_select if acl_select else ["internal"] * 8
        self.adc_channels = adc_channels if adc_channels else [None]*8
        
    def set_acl_select(self, acl_select):
        "Overwrite the internal/external selection. Do not use -- Use :meth:`Routing.front_input` instead."
        self.acl_select = acl_select
        
    def set_adc_channels(self, adc_channels):
        "Overwrite the ADC Channel requests. Do not use -- Use :meth:`Routing.probe` instead."
        self.adc_channels = adc_channels

    def measure(self, source:Union[Ele,int], adc_channel=None):
        """
        Syntactic sugar to set an adc_channel.
        """
        if adc_channel == None:
            # TODO is untested
            adc_channel = next_free(map(notNone, self.adc_channels))
        if not adc_channel in range(0,8):
            raise ValueError(f"{adc_channel=} illegal, expecting in 0..7")
        # first have to look for element in routing
        uin = source.out if isEle(source) else source
        self.adc_channels[adc_channel] = uin
        return adc_channel
        
    def generate(self):
        "Returns carrier configuration, not cluster configuration"
        ret = {}
        # TODO: ACL_SELECT is currently handled by the Routing class.
        #if self.acl_select:
            #if not all(isinstance(v, bool) for v in self.acl_select):
            #    raise ValueError(f"Unsuitable ACL selects, expected list of bools: {self.acl_select}")
        #    ret["acl_select"] = self.acl_select
        if len(list(filter(notNone, self.adc_channels))):
            # TODO: Probably check again with the Nones.
            if not all(isinstance(v, int) or v == None for v in self.adc_channels):
                raise ValueError(f"Unsuitable ADC channels, expected list of ints: {self.adc_channels}")
            ret["adc_channels"] = self.adc_channels
        return ret
    
    def load(self, config):
        # TODO: ACL_SELECT is currently handled by the Routing class.
        #if "acl_select" in config:
        #    self.acl_select = config["acl_select"]
        if "adc_channels" in config:
            self.adc_channels = config["adc_channels"]
        return self

#: Just a tuple collecting U, C, I matrices without determining their actual representation.
#: The representation could be lists, numpy arrays, etc. depending on the use.
UCI = namedtuple("UCI", ["U", "C", "I"])

class Routing(Probes):
    """
    This class provides a route-tuple like interface to the UCI block and
    generates the Output-centric matrix configuration at the end. Most likely,
    as a user you don't want to intiate a Routing instance but a Circuit instance
    instead.
    
    :arg accept_dirty: Flag for allowing to add "dirty" Routes, i.e. illegal Routes,
        without raising at :meth:`add`. Instead, you can use :meth:`sanity_check`
        later to see the problems once more. May be useful for importing existing
        circuits.
    """
    max_lanes = 32
    #routes : List[Route]
        
    def available_lanes(self):
        """
        Returns a list of lane indices generally available in the LUCIDAC (independent of
        their allocation). If you set the class attribute ``lanes_constraint``, their values
        will be used.
    
        >>> r = Routing()
        >>> r.lanes_constraint = [3, 7, 17]
        >>> r.connect(0, 8)
        Route(uin=0, lane=3, coeff=1, iout=8)
        >>> r.connect(3, 7)
        Route(uin=3, lane=7, coeff=1, iout=7)
        >>> r.connect(12, 14)
        Route(uin=12, lane=17, coeff=1, iout=14)
        >>> r.connect(7, 9)
        Traceback (most recent call last):
        ...
        ValueError: All [3, 7, 17] available lanes occupied, no more connections possible.

        """
        # for a fully functional lucidac, do this:
        #return list(range(32))
        # Instead, we know these lanes are working only:
        if hasattr(self, "lanes_constraint"):
            return self.lanes_constraint
        else:
            return list(range(32))
        #return [0,1,2,3,4,5, 14,15, 16,17, 18,19, 20,21, 30,31]
    
    def __repr__(self):
        return f"Routing({pprint.pformat(self.routes)})"
    
    def __init__(self, routes: List[Route] = None, accept_dirty=False, **kwargs):
        super().__init__(**kwargs)  # forwards all unused arguments
        self.routes = []
        self.u_constant = None
        self.accept_dirty = accept_dirty
        if routes:
            self.add(routes)
    
    def randomize(self, num_lanes=32, max_coeff=+10, seed=None):
        """
        Appends random routes.
    
        :arg num_lanes: How many lanes to fill up. By default fills up all lanes.
        :arg max_coeff: Maximal coefficient magnitudes. ``+1`` or ``+10`` are
             useful values for LUCIDAC.
        :arg seed: For reproducability (as in unit tests), this calls ``random.seed``.
             A large integer may be a suitable argument.
        :returns: The instance, i.e. is chainable.
        """
        if seed:
            random.seed(seed)
        for lane in range(num_lanes):
            uin, iout = random.randrange(16), random.randrange(16)
            coeff = random.uniform(-abs(max_coeff), +abs(max_coeff))
            self.add( Route(uin, lane, coeff, iout) )
        return self

    
    def use_constant(self, use_constant=True):
        """
        Activates/Configures the system's constant giver source. A constant giver allows to
        use constant numbers within the computation. A constant then can be connected for
        instance to an integrator or multiplier or summed with other values or constants.
        
        In LUCIDAC, there is one constant source in the U-Block which can configured with
        these values:
    
        * ``True`` or ``1.0`` or ``1``: Generates the constant ``+1``
        * ``0.1`` generates the constant ``+0.1``
        * ``False`` removes the overall constant
        
        Do not use ``None`` for turning off the constant, as this is the default value for
        not passing the constant request at all to the LUCIDAC.
        
        The constant can be further modified by the coefficient in the lane. This allows for
        (up and down) scaling and sign inversion of the constant.
        
        This method only registers the use of the constant. In order to make use of it, routes
        using Constants have to be added. As the constants are only available at certain
        clane/lane positions, you should make use of the :meth:`Reservoir.const` method in
        order to obtain a Constant object and connect it with :meth:`connect`, which will
        figure out a suitable lane.
        """
        self.u_constant = use_constant
    
    def probe(self, source:Union[Ele,int], front_port=None, weight=1):
        """
        Syntactic sugar to route a port to the front panel output.
        The name indicates that an oscilloscope probe shall be connected
        to this port.
        If no port is given, will count up.
        
        If you need a weight, use :meth:`connect` directly as in
        ``circuit.connect(source, circuit.front_output(front_port), weight=1.23)``.
       
        :arg front_output: Integer describing the front port number (0..7)
        :returns: The generated Route (is also added)
        
        See also :meth:`Circuit.measure` for putting a singal to ADC/DAQ.
        
        Instead of using this function, you can also add the route by yourself:
        """
        target = self.front_panel(front_port) # None passes to default alloc None!
        return self.add(Connection(source, target, weight=weight))
    
    def front_input(self, id, use_front=True):
        """
        Make use of a front panel input (or use internal lane instead, if `use_front=False`)
        Note that front panel ports are always outputs and that they come behind the C-block, thus
        no coefficient can be used.
        
        You should not use this function when describing circuits, since it won't give you control
        over the I-block configuration. Instead, you should route front inputs:
        
        >>> c = Circuit()
        >>> i0 = c.int() # integrator 0
        >>> fp0 = c.front_panel(0) # front panel input/output 0
        >>> c.connect(fp0, i0) # connect front panel input to integrator 0
        Route(uin=-1, lane=24, coeff=1, iout=0)
        """
        if id in range(DefaultLUCIDAC.acl_offset, DefaultLUCIDAC.acl_offset+DefaultLUCIDAC.num_acls):
            # also accept wrongly taken ids. Should ensure that there is no overlap between the two
            # accepted input domains.
            id -= DefaultLUCIDAC.acl_offset
        if not id in range(0,8):
            raise ValueError(f"{id=} must be within the range of front panel port number (0,8)")
        self.acl_select[id] = "external" if use_front else "internal"
    
    def next_free_lane(self, constraint=None):
        """
        Allocates and returns the next free lane in the circuit.
        
        :arg constraint: can be a callback which gets a candidate lane and can accept it as suitable
           or not. This is useful for instance for the constant sources or ACL_IN/OUT features
           which are only available on certain lanes or lane combinations.
        """
        route_for_lane = [ find(lambda r: r.lane == lane, None, self.routes) for lane in self.available_lanes() ]
        is_lane_occupied = [ True if x else False for x in route_for_lane ]
        #print("next_free_lane", self.available_lanes())
        #print("next_free_lane", is_lane_occupied)
        #occupied_lanes = [ r.lane for r in self.routes ]
        idx = next_free(is_lane_occupied, criterion=constraint)#, append_to=self.max_lanes)
        #print("next_free_lane = ",idx)
        if idx == None:
            raise ValueError(f"All {self.available_lanes()} available lanes occupied, no more connections possible." +\
                             ("" if not constraint else " But note that custom constraints were applied!"))
        return self.available_lanes()[idx]
    
    def add(self, route_or_list_of_routes:Union[Route,List[Route]]):
        """
        Main method for adding one or multiple routes to the internal route list.
        
        If a route containing Elements or a Connection (i.e. a Route with ``lane=None``) is given, this function
        will do an "immediate pick and place". So if you want, this is the *compiler* component in this class.
        If a "final" route is given, with only numeric arguments, there is not much to do.
        
        As a matter of principle, this function never *corrects* data it receives. That is, it *does* a 
        certain amount of proof/integrity checking which yields ``ValueErrors`` if we complain about:
        
        - Checking whether a Constant can be routed that way
        - Checking if the requested lane is already in use
        - Checking whether values are out of bounds (i.e. lanes, coefficients, etc)
        
        If you don't want this function to raise on invalid data, set the class attribute ``accept_dirty=True``.
        This option can also be passed to the constructor.
        """
        if isinstance(route_or_list_of_routes, list):
            return list(map(self.add, route_or_list_of_routes))
        
        old_route = route_or_list_of_routes
        uin, lane, coeff, iout = old_route
       
        if isinstance(uin, Const):
            right_ublock_chip  = lambda potential_lane: 15 < potential_lane
            left_ublock_chip   = lambda potential_lane: potential_lane < 16
            if lane is None:
                if uin.out == 14:
                    criterion = right_ublock_chip
                elif uin.out == 15:
                    criterion = left_ublock_chip
                else:
                    raise ValueError(f"Unacceptable Constant clane requested in unrouted {old_route}")
                lane = self.next_free_lane(criterion)
            else:
                if uin.out == 14 and not right_ublock_chip(lane):
                    raise ValueError(f"No Constant available at {old_route}, lane must be >15")
                if uin.out == 15 and not left_ublock_chip(lane):
                    raise ValueError(f"No Constant available at {old_route}, lane must be <16")
        
        acl_select = None
        
        if isinstance(uin, Front):
            self.front_input(uin.lane) # register input
            lane = uin.lane
            uin = Route.do_not_connect
            acl_select = "input"
            
        if isinstance(iout, Front):
            lane = iout.lane
            iout = Route.do_not_connect
            acl_select = "output"
        
        if lane is None:
            lane = self.next_free_lane()
            
        new_route = Route(uin, lane, coeff, iout)
    
        # at the end, replace the symbols with numbers.
        new_route.resolve()
    
        # out of bounds check
        if self.accept_dirty:
            for err in new_route.sanity_list():
                print(f"Warning at adding Route: {err}")
        else:
            new_route.sanity_raise()

        if acl_select:
            # search whether this lane was already used within the input/output ACL scheme
            # The idea is generally that within the current route,
            #    uin  == -1  corresponds to ACL_IN  request
            #    iout == -1  corresponds to ACL_OUT request
            # If we find a lane that has iout == -1 and we want ACL_IN, we can use that lane.
            # If we find a lane that has uin == -1 and we want ACL_OUT, we can use that lane
            for other in self.routes:
                if other.lane == lane:
                    other_acl_select = None
                    if other.uin == Route.do_not_connect:
                        other_acl_select = "input"
                    elif other.iout == Route.do_not_connect:
                        other_acl_select = "output"
                    else: # should not happen
                        raise ValueError(f"While trying to add a route, I found an allocated nonsense route {other}.")
                
                    if other_acl_select == acl_select:
                        raise ValueError(f"Cannot append {new_route} (external {acl_select}) because a similar one exists: {other}")
                    else: # can combine!
                        if other.uin == Route.do_not_connect:
                            other.uin = new_route.uin
                        elif other.iout == Route.do_not_connect:
                            other.iout = new_route.iout
                        else: # should not happen
                            raise ValueError(f"While trying to add a route, I caught me up in contradictions.")
                        # skip appending new_route because we merged it successfully into other
                        return other
            else: # have not found this ACL condition yet. Ensure this routing request makes sense:
                if acl_select == "input" and coeff != 1:
                    raise ValueError(f"Cannot add new ACL_INgoing {new_route} with a coefficient != 1 because it doesn't pass the C-block.")
                    # However, if user first adds an ACL_OUTgoing route at this lane, the ingoing will be accepted.
                    # But we have to ensure the weights match. Or the outgoing overwrites!
        else:
            if lane in [ r.lane for r in self.routes ]:
                raise ValueError(f"Cannot append {new_route} because this lane is already occupied.")
        
        self.routes.append(new_route)
        return new_route

    def connect(self, source:Union[Ele,int], target:Union[Ele,int], weight=1):
        """
        Syntactic sugar for adding a :func:`Connection`.
        :returns: The generated Route.
        """
        return self.add(Connection(source,target,weight))
    
    def route(self, uin, lane, coeff, iout):
        """
        Syntactic sugar for adding a :class:`Route`.
        :returns: The generated Route.
        """
        return self.add(Route(uin, lane, coeff, iout))
   
    def routes2input(self) -> UCI:
        """
        Converts the list of routes to an *input matrix representation*. In this format,
        the U/C/I matrix are each represented as a list of 32 numbers.
        This is basically an array-of-structures to structure-of-array (AoS-to-SoA) operation
        and inspired by the *lucicon* circuit compiler (part of `lucigui <https://github.com/anabrid/lucigui>`_).    
        See :meth:`generate` for the actual high-level method for the spare matrix format used
        by the LUCIDAC protocol.
    
        .. warning::
    
           The C matrix values in this representation are floats within ``[-20,+20]`` and not
           yet rescaled (as in REV1 hardware). See :meth:`coeff_upscale` for the code which shifts
           this information.
    
        """
        return UCI(
            U=clean([[r.uin  for r in self.routes if r.lane == lane and r.uin != Route.do_not_connect] for lane in range(32)]),
            I=clean([r.iout for r in self.routes if r.lane == lane and r.iout != Route.do_not_connect] for lane in range(32)),
            C=[route.coeff if route else 0 for route in (find(lambda r, lane=lane: r.lane == lane, None, self.routes) for lane in range(32))]
        )
    
    @staticmethod
    def input2output(inmat, keep_arrays=True):
        """
        Input/Output centric format conversion. Relevant only for C block.
        Maps ``Array<int|None,32>`` onto ``Array<Array<int>|int, 16>``.
        
        Example:
        
        >>> iouts_inputs = [ r.iout for r in Routing().randomize(seed=37311).routes ]
        >>> print(iouts_inputs)
        [7, 5, 2, 9, 2, 15, 7, 1, 5, 7, 6, 1, 5, 0, 10, 7, 13, 3, 9, 9, 10, 1, 9, 2, 1, 7, 12, 8, 10, 10, 1, 3]
        >>> Routing.input2output(iouts_inputs) # doctest: +NORMALIZE_WHITESPACE
        [[13],
         [7, 11, 21, 24, 30],
         [2, 4, 23],
         [17, 31],
         [],
         [1, 8, 12],
         [10],
         [0, 6, 9, 15, 25],
         [27],
         [3, 18, 19, 22],
         [14, 20, 28, 29],
         [],
         [26],
         [16],
         [],
         [5]]

        """
        output = [[] for _ in range(16)] # Array<Array, 16>
        for lane, clane in enumerate(inmat):
            if clane != None:
                if isinstance(clane, list):
                    for ci in clane:
                        output[ci].append(lane)
                else:
                    output[clane].append(lane)
        return output if keep_arrays else clean(output)
    
    @staticmethod
    def output2input(outmat):
        """
        Input/Output centric format conversion. Relevant only for C block.
        Maps ``Array<Array<int>|int,16>`` onto ``Array<int|None,32>``.
        
        This is obviously the inverse of ``input2output``:
        
        >>> iouts_inputs  = [ r.iout for r in Routing().randomize(seed=75164).routes ]
        >>> iouts_outputs = Routing.input2output(iouts_inputs)
        >>> again_inputs  = Routing.output2input(iouts_outputs)
        >>> iouts_inputs == again_inputs
        True
        
        :returns: List of size 32.
        """
        input = [ None for _ in range(32) ]
        for clane, lanes in enumerate(outmat):
            if lanes:
                for lane in lanes:
                    input[lane] = clane
        return input
    
    @staticmethod
    def coeff_upscale(c_elements, raise_out_of_bounds=True):
        upscaling = [ (v < -1 or v > 1) for v in c_elements ]
        scaled_c = [ (c/10 if sc else c) for sc, c in zip(upscaling, c_elements) ]
        return upscaling, scaled_c
    
    def sanity_check(self, also_print=True):
        """
        Performs a number of plausibilty and sanity checks on the given circuit. These are:
        
        1. General data type check on the given routes (out of bounds, etc)
        2. For computing elements with more then one input (currently only the multipliers):
           Checks whether no input or all are used.
        
        The sanity check is the last bastion between ill-defined data and a writeout to
        LUCIDAC, which will probably complain in a less comprehensive way.
        
        :returns: A list of human readable messages as strings. Empty list means no warning.
        
        Examples on errnous circuits:
        
        >>> a = Circuit()
        >>> a.routes.append(Route(17, -32, -24.0, None))
        >>> warnings_as_list = a.sanity_check()
        Sanity check warning: Route contains None values.
        Sanity check warning: Uin out of range in Route(uin=17, lane=-32, coeff=-24.0, iout=None)
        Sanity check warning: Lane out of range in Route(uin=17, lane=-32, coeff=-24.0, iout=None)
        Sanity check warning: Coefficient out of range in Route(uin=17, lane=-32, coeff=-24.0, iout=None)
        Sanity check warning: Iout out of range in Route(uin=17, lane=-32, coeff=-24.0, iout=None)
        
        Obviously, in the example above the problem started in the first place because the user
        accessed the ``routes`` atribute instead of using the :meth:`route` or :meth:`add`
        methods, which already do a good part of the checking. In general, never access the
        ``routes`` directly for writing.
        
        The situation is more difficult when at routing time the problem is not detectable but
        later it is:
        
        >>> b = Circuit()
        >>> i = b.int()
        >>> m = b.mul()
        >>> b.connect(i, m.a)
        Route(uin=0, lane=0, coeff=1, iout=8)
        >>> warnigns_as_list = b.sanity_check()
        Sanity check warning: Multiplier 0 (counting from 0) has input but output is not used (clane=8 does not go to some route.uin).
        Sanity check warning: Warning: Multiplier 0 is in use but connection B is empty

        There are a number of mis-uses which this checker cannot detect, by design. This
        is, for example, when inputs and outputs are mixed up:
        
        >>> c = Circuit()
        >>> correct0 = c.connect(i, m.a)
        >>> correct1 = c.connect(i, m.b)
        >>> errnous = c.connect(m.a, i.out)
        >>> c.sanity_check() == []
        True

        The reason for this is because in this example, the explicit member access ``m.a`` or
        ``i.out`` resolves to integers and the checker cannot find out whether the indices where
        given intentionally or by accident.


        """
        warnings = []
        
        ### General route check
        # These are actually not warnings but errors.
        for route in self.routes:
            warnings += route.sanity_list()
        
        ### Multiplier check
        multipliers_used = [False]*DefaultLUCIDAC.num_mul
        multipliers = [ DefaultLUCIDAC.make(Mul, i) for i in range(DefaultLUCIDAC.num_mul) ]

        # first determine which ones have either inputs or outputs connected
        for route in self.routes:
            for m in multipliers:
                if route.uin == m.out or route.iout == m.a or route.iout == m.b:
                    multipliers_used[m.id] = True
                    
        def has_connection(test):
            connected = False
            for route in self.routes:
                if test(route):
                    connected = True
            return connected
        
        # then determine if all ports are connected
        for i,(used, m) in enumerate(zip(multipliers_used, multipliers)):
            if used:
                if not has_connection(lambda route: route.uin == m.out):
                    warnings.append(f"Multiplier {i} (counting from 0) has input but output is not used (clane={m.out} does not go to some route.uin).")
                if not has_connection(lambda route: route.iout == m.a):
                    warnings.append(f"Warning: Multiplier {i} is in use but connection A is empty")
                if not has_connection(lambda route: route.iout == m.b):
                    warnings.append(f"Warning: Multiplier {i} is in use but connection B is empty")

        # TODO should use logging instead
        if also_print:
            for warning in warnings:
                print(f"Sanity check warning: {warning}")

        return warnings

    def generate(self, sanity_check=True):
        """
        Generate the configuration data structure required by the JSON protocol, which is
        that "output-centric configuration". This is a major function of the class. You
        can use the data structure generated here immediately to produce JSON configuration
        files as standalone, for instance by passing the output of this function to
        ``json.dumps(circuit.generate())``.
        
        Generates configuration for the Carrier.
        
        .. note::
        
           The C-Matrix values here are correctly scaled.
        """
        
        if sanity_check:
            self.sanity_check()
        
        U,C,I = self.routes2input()
        upscaling, scaled_c = self.coeff_upscale(C)
        
        cluster = {
            "/U": dict(outputs = U),
            "/C": dict(elements = scaled_c),
            "/I": dict(outputs = self.input2output(I), upscaling = upscaling)
        }
        
        # hopefully this also allows to turn OFF the constant by setting it to False...
        if self.u_constant != None:
            cluster["/U"]["constant"] = self.u_constant
            
        carrier = {"/0": cluster}

        carrier.update(Probes.generate(self).items())
        
        if any(k == "external" for k in self.acl_select):
            carrier["acl_select"] = self.acl_select
        
        return carrier
        
    def load(self, carrier_config):
        """
        The inverse of generate. Useful for loading an existing circuit. You can load an
        existing standalone JSON configuration file by using this function.
        
        The following code snippet shows that :meth:`load` and :meth:`generate` are really
        inverse to each other, allowing to first export and then import the same circuit:
        
        >>> a = Routing().randomize()
        >>> b = Routing().load(a.generate(sanity_check=False))
        >>> a.routes == b.routes
        True
        
        .. note::
        
           Also accepts a carrier-level configuration. Does not accept a configuration message
           including the JSON Envelope right now.
            
            Currently does not yet properly reads ACL_IN requests.
        """
        
        # load parental classes
        Probes.load(self, carrier_config)
        
        # also accept slightly illegal input
        cluster_config = carrier_config["/0"] if "/0" in carrier_config else carrier_config
        
        # load all in output centric format
        U = cluster_config["/U"]["outputs"] if "/U" in cluster_config else [ None ]*32
        C = cluster_config["/C"]["elements"] if "/C" in cluster_config else [ 0.0  ]*32
        I = cluster_config["/I"]["outputs"] if "/I" in cluster_config else [ None ]*16
        I = self.output2input(I)
        
        if "upscaling" in cluster_config["/I"]:
            for i in range(32):
                if cluster_config["/I"]["upscaling"][i]:
                    C[i] *= 10
        
        if "constant" in cluster_config["/U"]:
            self.use_constant(cluster_config["/U"]["constant"])
        
        for lane,(u,c,i) in enumerate(zip(U,C,I)):
            if u == None:
                continue
            if i == None or i == []:
                continue
            if not isinstance(i, list):
                i = [i]
            for ii in i:
                self.add(Route(u,lane,c,ii))
        return self
    
    def to_pybrid_cli(self):
        """
        Generate the Pybrid-CLI commands as string out of this Route representation.
        """
        self.sanity_check()
        return "\n".join(f"route -- carrier/0 {r.uin:2d} {r.lane:2d} {r.coeff: 7.3f} {r.iout:2d}" for r in self.routes)
    
    def to_dense_matrix(self, sanity_check=True):
        """
        Generates a dense numpy matrix for the UCI block, i.e. a real-valued 16x16 matrix with
        bounded values [-20,20] where at most 32 entries are non-zero.
        """
        if sanity_check:
            self.sanity_check()
        import numpy as np
        UCI = np.zeros((16,16))
        for (uin, _lane, coeff, iout) in self.routes:
            if iout != Route.do_not_connect:
                UCI[iout,uin] += coeff
        return UCI
    
    def to_dense_matrices(self, sanity_check=True) -> UCI:
        """
        Generates the three matrices U, C, I as dense numpy matrices.
        C and I are properly scaled as in the real system (upscaling happens in I).
        
        :returns: 3-Tuple of numpy matrices for U, C, I, in this order.
        
        Note that one way to reproduce :meth:`to_dense_matrix` is just by
        computing ``I.dot(C.dot(U))`` on the output of this function.
        
        >>> import numpy as np
        >>> c = Circuit().randomize()
        >>> U, C, I = c.to_dense_matrices(sanity_check=False) # skipping for doctesting
        >>> np.all(I.dot(C.dot(U)) == c.to_dense_matrix(sanity_check=False))
        True
        
        Limitations: This omits all information about ``acl_select`` and the constant
        giver scheme.
        """
        if sanity_check:
            self.sanity_check()
        import numpy as np
        U = np.zeros((32,16))
        C = np.zeros((32,32))
        I = np.zeros((16,32))
        
        for (uin, lane, coeff, iout) in self.routes:
            assert not None in (uin, lane, iout), "Nones behave badly in numpy array subscription. Sanatize your routes."
            U[lane, uin]  = 1
            I[iout, lane] = 1 if abs(coeff) < 10 else 10
            C[lane, lane] = coeff / I[iout, lane]
            if iout == Route.do_not_connect:
                I[iout, lane] = 0
        
        return UCI(U,C,I)

    def to_sympy(self, int_names=None, subst_mul=False, no_func_t=False):
        """
        Creates an ODE system in sympy based on the circuit.
        
        :arg int_names: Allows to overwrite the names of the integators. By default they
           are called i_0, i_1, ... i_7. By providing a list such as ["x", "y", "z"], these
           names will be taken. If the list is shorter then length 8, it will be filled up
           with the default names.
        :arg subst_mul: Substitute multiplications, getting rid of explicit Eq(m_0, ...)
           statements. Useful for using the equation set within an ODE solver because the
           state vector is exactly the same length of the entries.
        :arg no_func_t: Write f instead of f(t) on the RHS, allowing for denser notations,
           helps also in some solver contexts.
        
        .. warning ::
        
           This method is part of the Routing class and therefore knows nothing about
           the MIntBlock settings, in particular not the k0. You have to apply different
           k0 values by yourself if neccessary! This can be as easy as to multiply the
           equations rhs with the appropriate k0 factor.
        
        Further limitations:
        
        * Will just ignore ACL_IN (Front panel in/out)
        
        Example for making use of the output within a scipy solver:
        
        ::
        
            from sympy import latex, lambdify, symbols
            from scipy.integrate import solve_ivp
            xyz = list("xyz")
            eqs = lucidac_circuit.to_sympy(xyz, subst_mul=True, no_func_t=True)
            print(eqs)
            print(latex(eqs))
            rhs = [ e.rhs for e in eqs ]
            x,y,z = symbols(xyz)
            f = lambdify((x,y,z), rhs)
            f(1,2,3) # works
            sol = solve_ivp(f, (0, 10), [1,2,3])
        
        Limitations: This omits all information about ``acl_select`` and the constant
        giver scheme.
        """
        self.sanity_check()
        from sympy import symbols, Symbol, Function, Derivative, Eq
        
        t = symbols("t")
        
        int_default_names = [ f"i_{d}" for d in range(8) ]
        if int_names:
            for i,n in enumerate(int_names):
                int_default_names[i] = n
        
        ints = [Function(n)(t) for n in int_default_names]
        muls = [Function(f"m_{d}")(t) for d in range(4)]
        sympy_reservoir = { Int: ints, Mul: muls, Const: [1.0]*4 }
        di = [ Derivative(i, t) for i in ints ]
        
        sympy_uin = [ DefaultLUCIDAC.resolve_mout(route.uin, sympy_reservoir) for route in self.routes ]
        #summands = [ coeff * DefaultLUCIDAC.resolve_mout(uin, sympy_reservoir) for (uin, _lane, coeff, iout) in self.routes ]
        min_sums = [ sum(route.coeff*uin_symbol for uin_symbol, route in zip(sympy_uin, self.routes) if route.iout == min) for min in range(16) ]
        
        min_inputs = itertools.batched(range(0,8), 2) # [(0, 1), (2, 3), (4, 5), (6, 7)]
        
        # these will contain still-to-filter entries like Eq(m_3(t),0)
        mul_eqs = [ Eq(mi, -(min_sums[input_indices[0]] * min_sums[input_indices[1]])) for mi, input_indices in zip(muls, min_inputs) ]
        int_eqs = [ Eq(di[idx], min_sums[idx+DefaultLUCIDAC.MIntOffset]) for idx in range(0,8) ]
        
        if subst_mul:
            mappings = { eq.lhs: eq.rhs for eq in mul_eqs }
            # replace recursive calls such as in m0=x*x, my=x*m0
            mappings = { eq.lhs: eq.rhs.subs(mappings) for eq in mul_eqs }
            # then apply on int definitions
            eqs = [ eq.subs(mappings) for eq in int_eqs ] 
        else:
            eqs = mul_eqs + int_eqs
            
        if no_func_t:
            mappings = { Function(n)(t): Symbol(n) for n in int_default_names }
            eqs = [ eq.subs(mappings) for eq in eqs]
            
        return [ eq for eq in eqs if eq.rhs != 0 ]


    def reverse(self):
        """
        Trivially "reverse engineer" a circuit based on routes. Tries to output valid
        python code as string.
        
        Limitations: This omits all information about ``acl_select`` and the constant
        giver scheme.
        """
        self.sanity_check()
        
        # TODO: Code is ugly, refactor to proper Int/Mul types which can do most of this
        
        def assert_one(candidates, r):
            candidates = list(candidates)
            if len(candidates) == 0:
                raise ValueError(f"Route {r} not assignable, no candidates")
            if len(candidates) == 1:
                return candidates[0]
            if len(candidates) == 2:
                raise ValueError(f"Route {r} overassignable, candidates are {src_candidates}")
        
        populated = DefaultLUCIDAC.populated()
        
        def within(itm, fields, idx):
            for n in fields:
                if hasattr(itm, n) and getattr(itm, n) == idx:
                    return n
            return None

        def route2connection(r):
            # TODO Warning, the target Mul.a|b is most likely still wrong. Has to be tested,
            #      see for instance roessler or neuron as an errnous example.
            
            #print("source ", r)
            source = assert_one(filter(lambda itm: itm.out == r.uin, populated), r)
            #print("target ", r)
            target = assert_one(filter(lambda itm: bool(within(itm, ["a", "b"], r.iout)), populated), r)
            target_port = assert_one(filter(bool, [ within(itm, ["a", "b"], r.uin) for itm in populated ]), r)
            has_two_ports = hasattr(target, "b") # i.e. if it is Mul or Not
            weight = r.coeff
            source_shortname = type(source).__name__ + str(source.id)
            target_shortname = type(target).__name__ + str(target.id) + ("."+target_port if has_two_ports else "")
            return f"Connection(" + source_shortname + ", " + target_shortname + (f", {weight=}" if r.coeff != 1 else "") + ")"
                         
        return ",\n".join(map(route2connection, self.routes))
        
class Circuit(Reservoir, MIntBlock, Routing):
    """
    The Circuit class collects all reconfigurable properties of a LUCIDAC. The class joins
    all features from it's parents in a *mixin idiom*. (This means that, if you really want to,
    you also can use the individual features on its own. But there should be no need to do so)
        
    This class provides, for instance, an improved version of the 
    :meth:`Reservoir.int` method which also sets the integrator state (initial value and time factors)
    in one method call.

    Most of all, this function generates the final configuration format required for LUCIDAC.
    
    However, the mixin idiom also has it's limitations. For instance, for the ``Reservoir``, there
    will be no "backwards syncing", i.e. if routes are
    added manually without using the Reservoir, it's bookkeeping is no more working. Example:
        
    >>> circ = Circuit()
    >>> circ.connect(0, 0) # defacto connects M0 first output to M0 first input
    Route(uin=0, lane=0, coeff=1, iout=0)
    >>> I = circ.int()
    >>> print(I)
    Int(id=0, out=0, a=0)
    >>> circ.connect(I, I)
    Route(uin=0, lane=1, coeff=1, iout=0)

    One could expect in this example that ``circ.int()`` hands out the second integrator (``id=1``)
    but it does not. The registry only knows about how often it was called and does not know at
    all how the computing elements are used in the Routing.
    """
    
    def __init__(self, routes: List[Route] = [], accept_dirty=False):
        super().__init__(routes=routes, accept_dirty=accept_dirty)
    
    def int(self, *, id=None, ic=0, slow=False):
        "Allocate an Integrator and set it's initial conditions and k0 factor at the same time."
        el = Reservoir.int(self, id)
        self.set_ic(el, ic)
        self.set_k0(el, MIntBlock.slow if slow else MIntBlock.fast)
        return el
    
    def load(self, config_message):
        """
        Loads the configuration which :meth:`generate` spills out. The full circle.
        """
        config_carrier = config_message["config"] if "config" in config_message else config_message
        config_cluster = config_carrier["/0"] if "/0" in config_carrier else config_carrier
        
        if "/M0" in config_cluster:
            MIntBlock.load(self, config_cluster["/M0"])
        Routing.load(self, config_carrier)

        return self
    
    def randomize(self, num_lanes=32, max_coeff=+10, seed=None):
        """
        Add random configurations. This function is basically only for
        testing. See :meth:`Routing.randomize` for the paramters.
    
        :arg num_lanes: How many lanes to fill up. By default fills up all lanes.
        :arg max_coeff: Maximal coefficient magnitudes. ``+1`` or ``+10`` are
             useful values for LUCIDAC.
        :arg seed: For reproducability (as in unit tests), this calls ``random.seed``.
             A large integer may be a suitable argument.
        :returns: The instance, i.e. is chainable.
        """
        MIntBlock.randomize(self)
        Routing.randomize(self, num_lanes, max_coeff, seed)
        return self
    
    def generate(self, skip=None, sanity_check=True):
        """
        Returns the data structure required by the LUCIDAC set_config call *for a given carrier*,
        not cluster. The output of this function can be straightforwardly fed into ``LUCIDAC.set_circuit``.
        
        The LUCIDAC cluster is always part of the carrier level configuration:
        
        >>> "/0" in Circuit().randomize().generate(sanity_check=False)
        True
        
        :arg skip: An entity (within LUCIDAC, at Cluster level) to skip, for instance "/M1"
        :arg sanity_check: Whether to carry out a sanity check (results in printouts)
        """
        carrier_config = Routing.generate(self, sanity_check=sanity_check)
        carrier_config["/0"]["/M0"] = MIntBlock.generate(self) 
        carrier_config["/0"]["/M1"] = {} # MMulBlock
        
        if skip:
            carrier_config["/0"] = { k:v for k,v in carrier_config["/0"].items() if not skip in k }

        return carrier_config
        
    def write(self, hc, **args):
        "Shorthand to write out configuration to hybrid controller"
        hc.set_circuit(self.generate(**args))

    def to_json(self, **args):
        "Shorthand to get JSON. Typically you don't need this"
        import json
        config = self.generate(**args)
        return json.dumps(config)
        
    def to_ascii_art(self, full_Cblock=False):
        """
        Creates an "ASCII art" of the LUCIDAC including the current configuration.
        
        Includes: U, C, I, Mblock configuration
        
        Does not yet include:
        
        - ACL IN/OUT
        - Constant givers
      
        """
        
        # The implementation/code for this method stems from the C libsim and thus may look
        # a bit awkward in python.
        
        import numpy as np
        U, C, I = self.to_dense_matrices()
        U = U.T
            
        dump = "LuciPy LUCIDAC ASCII Dump      +--[ UBlock ]----------------------+\n" + \
               "+-[ M0 = MIntBlock ]-----+     | 0123456789abcdef0123456789abcdef |\n"
            
        for i in range(8):
            log10k0 = np.log10(self.k0s[i])
            assert log10k0 == int(log10k0)
            dump += "| INT%ld IC=%+.2f  k0=10^%d |" % (i, self.ics[i], log10k0);
            u_clane_used = np.any(U[i,:])
            c = "A" if i in self.adc_channels else ("-" if u_clane_used else ".")
            dump += f" -{c}> " if u_clane_used else f"  {c}  "
            dump += "%lX " % i
            
            # U Block upper
            dump += "".join("X" if U[i,j] else "." for j in range(32))
            
            dump += " %lX\n" % i
            
        dump += "+-[ M1 =  MulBlock ]-----+\n"
            
        for i in range(8):
            muli, ab = "MUL%ld" % (i/2), "a" if i%2==0 else "b"
            if i < 4:
                dump += "| %s.%s        MUL%ld.out |" % (muli, ab, i)
            else:
                dump += "| %s.%s                 |" % (muli, ab)
            
            # U Block lower
            u_clane_used = np.any(U[i+8,:])
            c = "A" if i in self.adc_channels else ("-" if u_clane_used else ".")
            dump += f" -{c}> " if u_clane_used else f"  {c}  "
            dump += "%lX " % (i+8)
            dump += "".join("X" if U[i+8,j] else "." for j in range(32))
            dump += " %lX\n" % (i+8)

        Uout1 = "".join("v" if np.any(U[:,j]) else "-" for j in range(32))
        #Uout2 = "".join("v" if np.any(U[:,j]) else "" for j in range(32))

        dump += f"+------------------------+     +-{Uout1}-+\n";
        #dump += f"                                  {Uout2}\n";
        dump += "\n"
        
        dump += "                               +--[ CBlock ]----------------------+\n" + \
                "                               | 0123456789abcdef0123456789abcdef |\n"
            
        # Dumping the c block (up to 80 chars width)
        Cwhite = " "*34 # place of M blocks
        dump += Cwhite + "\n"
        #dump += "".join("v" if C[i,i] else " " for i in range(32))

        for i in range(32):
            factor = C[i,i]
            if not full_Cblock and factor == 0:
                continue
            dump += Cwhite
            dump += "".join(("X" if j==i else ("|" if np.diag(C)[j]!=0 else ".")) if factor else "." for j in range(32) )
            dump += " C%02ld = %+02.3f " % (i, factor)
            dump += "\n"
            
        dump += Cwhite
        dump += " "*32; # cblock->factor[i] ? "|" : " ";
        dump += "\n" #dump += "+--------------+\n"
            
            
        dump += "                               +--[ IBlock ]----------------------+\n" + \
                "+-[ M0 = MIntBlock ]-----+     | 0123456789abcdef0123456789abcdef |\n"
            
        for i in range(8):
            # M0 block
            dump += "|                   INT%ld |" % i
            dump += " <-- " if np.any(I[i,:]) else "  .  "
            dump += "%lX " % i
            
            # I block upper
            dump += "".join("X" if I[i,j] else "." for j in range(32))
            dump += " %lX\n" % i
            
        dump += "+-[ M1 =  MulBlock ]-----+\n"
            
        for i in range(8):
            # M1 block
            muli, ab = "MUL%ld" % (i/2),  "a" if i%2==0 else "b"
            dump += "|                 %s.%s |" % (muli, ab)

            # I Block lower
            dump += " <-- " if np.any(I[i+8,:]) else "  .  "
            dump += "%lX " % (i+8)
            dump += "".join("X" if I[i+8,j] else "." for j in range(32))
            dump += " %lX\n" % (i+8)
        
        dump += "+------------------------+     +----------------------------------+\n"

        return dump
    
        
    def to_pybrid_cli(self):
        "Pybrid code generation including both the MIntBlock and Routing."
        nl = "\n"
        ret = "set-alias * carrier" + nl*2
        ret += MIntBlock.to_pybrid_cli(self) + nl
        
        ret += Routing.to_pybrid_cli(self) + nl
        
        ret += "# run --op-time 500000" + nl
        return ret



import pytest, numpy as np
from lucipy import LUCIDAC, Emulation, Circuit, Route

from fixture_circuits import circuit_sinus, measure_ramp

@pytest.fixture
def endpoint():
    emu = Emulation("127.0.0.1", 0)
    proc = emu.serve_forking()
    endpoint = emu.endpoint()
    yield endpoint
    proc.terminate()

def local_emulator():
    hc = Emulation()
    
    hc.mac = "foobar"
    hc.reset_circuit()
    return hc

def test_local_mac():
    hc = local_emulator()
    # Without going over TCP, the call returns the whole message
    assert list(hc.get_entities()["entities"].keys())[0] == hc.mac
    
def test_mac(endpoint):
    remote = LUCIDAC(endpoint)
    # The LUCIDAC API unwraps the message, returns only the useful payload
    assert list(remote.get_entities().keys())[0]  == Emulation().mac

def test_local_config():
    hc = local_emulator()
    config = circuit_sinus().generate()
    
    hc.set_circuit([hc.mac, "/0"], config)
    assert hc.get_circuit()["config"]["/0"] == config


def test_set_circuit_cluster(endpoint):
    hc = LUCIDAC(endpoint)
    set_conf_carrier = circuit_sinus().generate()
    hc.set_circuit(set_conf_carrier)
    get_conf_carrier = hc.get_circuit()["config"]

    assert "/0" in set_conf_carrier
    set_conf_cluster = set_conf_carrier["/0"]
    
    assert "/0" in get_conf_carrier
    get_conf_cluster = get_conf_carrier["/0"]
    
    print(f"{set_conf_cluster=}")
    print(f"{get_conf_cluster=}")
    assert set_conf_cluster == get_conf_cluster

def test_set_adc_channels(endpoint):
    hc = LUCIDAC(endpoint)
    c = Circuit()
    c.set_adc_channels([0,1,2])
    set_conf = c.generate()
    print(f"{set_conf=}")
    hc.set_config(set_conf)
    get_conf = hc.get_circuit()["config"]
    print(f"{get_conf=}")
    assert get_conf["adc_channels"] == c.adc_channels

def test_run_daq(endpoint):
    """
    This is a standalone test which makes sure the emulated data aquisition
    spills out the correct number of samples. This basically emulates how the
    FlexIO/DMA code should work. It also tests whether the result matches what
    is expected.
    """
    hc = LUCIDAC(endpoint)
    hc.reset_circuit()
    
    sinus = Circuit()
    
    k0 = 10_000 # system timescale factor if slow=False
    x = sinus.int(ic=+1, slow=False)
    y = sinus.int(ic=0, slow=False)
    
    sinus.connect(x, y)
    sinus.connect(y, x, weight=-1)
    
    sinus.measure(x)
    sinus.measure(y)
    channels = 2
    
    hc.set_circuit(sinus.generate())
    
    # this will just show roughly 1.5 wavelengths within the simulation
    t_final_ns = 900_000
    t_final_sec = t_final_ns / 1e9
    points_per_sec = 125_000 # something smallish for a quick result
    delta_t = 1./points_per_sec
    num_points = int(t_final_sec / delta_t)
    
    hc.set_daq(num_channels=2, sample_rate=points_per_sec)
    hc.set_run(op_time=t_final_ns)

    run = hc.start_run()
    data = np.array(run.data())
    
    assert data.shape == (num_points, channels)
    x_measured, y_measured = data.T
    
    t = np.linspace(0, t_final_sec, num_points)
    assert len(data) == len(t)

    # analytical solution to test problem:
    x_expected = -np.cos(t * k0) # corresponds to ic=+1
    y_expected = +np.sin(t * k0) # corresponds to ic=0
    
    assert np.allclose(x_expected, x_measured, atol=1e-2)
    assert np.allclose(y_expected, y_measured, atol=1e-2)


def test_ramp(endpoint):
    # Probably use @pytest.mark.parametrize instead to spawn "more tests"
    # and thus be faster
    hc = LUCIDAC(endpoint)
    hc.reset_circuit()

    slopes   = [-1,-0.5, 0, +0.5,+1]
    slopes  += [-10, -5, +5, +10]
    lanes    = [0,7,15,16,27,31] # no need to test range(0,32) in an emulator!
    
    slow = False

    res = []
    for lane in lanes:
        for slope in slopes:
            hc.reset_circuit()
            measure_ramp(hc, slope, lane, const_value=-1, slow=slow, do_assert=True)
        



#!/usr/bin/env python3

import setuptools, subprocess

exec = lambda cmd: subprocess.check_output(cmd.split()).decode().strip()

def get_version():
    try:
        ver = exec("git describe --tags --long") # something like "v0.1.0-1234-g123456"
        vABC, commits, githash = ver.split("-")
        versionlist = vABC.split(".") # major includes the "v"
        major, minor = versionlist[0], versionlist[1]
        patch = commits
        version = f"{major}.{minor}.{patch}"
        return version
    except (subprocess.CalledProcessError, FileNotFoundError): # no git installed
        return "v0.1.1234"

setuptools.setup(
    name="lucipy",
    version=get_version(),
)
